# -*- coding: utf-8 -*-
"""
.. module:: utils
    :synopsis: utils for basic functions

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """
import sys
import pandas as pd
import numpy as np
import requests
import math
import scipy.stats as spstats

from numba import jit

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

def groupby_count(df, colgroupby, colcountby, count_unique=True, show_progress=False):
    """
    Group the DataFrame and count the number for each group.

    Parameters
    ----------
    df: DataFrame
        The DataFrame.

    colgroupby: str
        The column to groupby.

    colcountby: str
        The column to count.

    count_unique: bool, default True
        If True, count unique items in the rows.  If False, just return the number of rows.

    show_progress: bool or str, default False
        If True, display a progress bar for the count.  If str, the name of the progress bar to display.

    Returns
    ----------
    DataFrame
        DataFrame with two columns: colgroupby, colcountby+`Count`
    """
    
    newname_dict = zip2dict([str(colcountby), '0'], [str(colcountby)+'Count']*2)

    if not show_progress:
        if count_unique:
            count = df.groupby(colgroupby, sort=False, as_index=False)[colcountby].nunique()
        else:
            count = df.groupby(colgroupby, sort=False, as_index=False)[colcountby].count()

    else:
        desc = ''
        if isinstance(show_progress, str):
            desc = show_progress
        # register our pandas apply with tqdm for a progress bar
        tqdm.pandas(desc=desc, disable= not show_progress)

        
        if count_unique:
            count = df.groupby(colgroupby, sort=False, as_index=False)[colcountby].progress_apply(lambda x: x.nunique())
        else:
            count = df.groupby(colgroupby, sort=False, as_index=False)[colcountby].progress_apply(lambda x: x.shape[0])

    return count.rename(columns=newname_dict)

def groupby_range(df, colgroupby, colrange, show_progress=False):
    """
   Group the DataFrame and find the range between the smallest and largest value for each group.

    Parameters
    ----------
    df: DataFrame
        The DataFrame.

    colgroupby: str
        The column to groupby.

    colrange: str
        The column to find the range of values.

    show_progress: bool or str, default False
        If True, display a progress bar for the range.  If str, the name of the progress bar to display.

    Returns
    ----------
    DataFrame
        DataFrame with two columns: colgroupby, colrange+`Range`
    """
    desc = ''
    if isinstance(show_progress, str):
        desc = show_progress
    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc=desc, disable= not show_progress)

    newname_dict = zip2dict([str(colrange), '0'], [str(colrange)+'Range']*2)
    return df.groupby(colgroupby, sort=False)[colrange].progress_apply(lambda x: x.max() - x.min()).to_frame().reset_index().rename(columns=newname_dict)

def groupby_zero_col(df, colgroupby, colrange, show_progress=False):
    """
    Group the DataFrame and shift the column so the minimum value is 0.

    Parameters
    ----------
    df: DataFrame
        The DataFrame.

    colgroupby: str
        The column to groupby.

    colrange: str
        The column to find the range of values.

    show_progress: bool or str, default False
        If True, display a progress bar.  If str, the name of the progress bar to display.

    Returns
    ----------
    DataFrame
        DataFrame with two columns: colgroupby, colrange
    """
    desc = ''
    if isinstance(show_progress, str):
        desc = show_progress
    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc=desc, disable= not show_progress)

    return df.groupby(colgroupby, sort=False)[colrange].progress_transform(lambda x: x - x.min())

def groupby_total(df, colgroupby, colcountby, show_progress=False):
    """
    Group the DataFrame and find the total of the column.

    Parameters
    ----------
    df: DataFrame
        The DataFrame.

    colgroupby: str
        The column to groupby.

    colcountby: str
        The column to find the total of values.

    show_progress: bool or str, default False
        If True, display a progress bar for the summation.  If str, the name of the progress bar to display.

    Returns
    ----------
    DataFrame
        DataFrame with two columns: colgroupby, colcountby+'Total'
    """
    desc = ''
    if isinstance(show_progress, str):
        desc = show_progress
    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc=desc, disable= not show_progress)

    newname_dict = zip2dict([str(colcountby), '0'], [str(colcountby)+'Total']*2)
    return df.groupby(colgroupby, sort=False)[colcountby].progress_apply(lambda x: x.sum()).to_frame().reset_index().rename(columns=newname_dict)

def groupby_mean(df, colgroupby, colcountby, show_progress=False):
    """
    Group the DataFrame and find the mean of the column.

    Parameters
    ----------
    df: DataFrame
        The DataFrame.

    colgroupby: str
        The column to groupby.

    colcountby: str
        The column to find the mean of values.

    show_progress: bool or str, default False
        If True, display a progress bar for the summation.  If str, the name of the progress bar to display.

    Returns
    ----------
    DataFrame
        DataFrame with two columns: colgroupby, colcountby+'Mean'
    """
    desc = ''
    if isinstance(show_progress, str):
        desc = show_progress
    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc=desc, disable= not show_progress)

    newname_dict = zip2dict([str(colcountby), '0'], [str(colcountby)+'Mean']*2)
    return df.groupby(colgroupby, sort=False)[colcountby].progress_apply(lambda x: x.mean()).to_frame().reset_index().rename(columns=newname_dict)

def isin_range(values2check, min_value, max_value):
    """
    Check if the values2check are in the inclusive range [min_value, max_value].

    Parameters
    ----------
    values2check: numpy array
        The values to check.

    min_value: float
        The lowest value of the range.

    max_value: float
        The highest value of the range.

    show_progress: bool or str, default False
        If True, display a progress bar for the count.  If str, the name of the progress bar to display.

    Returns
    ----------
    Numpy Array
        True if the value is in the range.
    """
    return np.logical_and(values2check >= min_value, values2check <= max_value)

def isin_sorted(values2check, masterlist):
    """
    Check if the values2check are in the sorted masterlist.

    Parameters
    ----------
    values2check: numpy array
        The values to check.

    masterlist: numpy array
        The sorted list of master values.

    Returns
    ----------
    Numpy Array
        True if the value is in the masterlist.
    """
    index = np.searchsorted(masterlist, values2check, side = 'left')
    index[index >= masterlist.shape[0]] = masterlist.shape[0] - 1
    return values2check == masterlist[index]

def argtopk(a, k=5):
    return np.argpartition(a, -k)[-k:][::-1]

def changepoint(a):
    return np.concatenate([[0], np.where(a[:-1] != a[1:])[0] + 1, [a.shape[0]]])

def holm_correction(pvalues, alpha=0.05):
    """
    Holm’s Step-Down Procedure for multiple hypothesis testing.
    Rather than controlling the FMER, Holm’s procedure controls for the false discovery rate (FDR)

    Parameters
    ----------
    pvalues: numpy array
        The p-values to check.

    alpha: float, defulat 0.05
        The significance level.

    Returns
    ----------
    Numpy Array
        True if the p-value is family-wise significant.

    """
    K = pvalues.shape[0]
    sorted_pvalues = np.sort(pvalues)
    holm_stepdown = alpha / (K - np.arange(K))
    
    first_reject_idx = np.argmax(sorted_pvalues > holm_stepdown)
    
    return np.argsort(pvalues) < first_reject_idx

def zip2dict(keys, values):
    return dict(zip(keys, values))

def series2df(ser):
    return ser.to_frame().reset_index()

def check4columns(df, column_list):
    # TODO: make proper error messages
    for col in column_list:
        if not col in list(df):
            raise KeyError("{} not in dataframe".format(col))

def pandas_cosine_similarity(df1, df2, col_keys, col_values):
    search_index = np.searchsorted(df2[col_keys].values, df1[col_keys].values, side = 'left')
    search_index[search_index >= df2.shape[0]] = df2.shape[0] - 1
    isin = df1[col_keys].values == df2[col_keys].values[search_index]
    cosine_num = np.inner(df1[col_values].values[isin], df2[col_values].values[search_index[isin]])
    cosine_denom = np.linalg.norm(df1[col_values].values) * np.linalg.norm(df2[col_values].values)
    return 1.0 - cosine_num/cosine_denom
    

def rolling_window(a, window, step_size = 1):
    shape = a.shape[:-1] + (a.shape[-1] - window + 1 - step_size, window)
    strides = a.strides + (a.strides[-1] * step_size,)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)

def forward_rolling_window(a, window, step_size = 1):
    a = np.pad(a.astype(float), (0,int(window)), 'constant', constant_values=(np.nan, np.nan))
    shape = a.shape[:-1] + (a.shape[-1] - window + 1 - step_size, window)
    strides = a.strides + (a.strides[-1] * step_size,)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)

def hard_rolling_window(a, window, step_size = 1):
    a = np.pad(a.astype(float), (int((window - 1)/2),int((window - 1)/2) + 1), 'constant', constant_values=(np.nan, np.nan))
    shape = a.shape[:-1] + (a.shape[-1] - window + 1 - step_size, window)
    strides = a.strides + (a.strides[-1] * step_size,)
    return  np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)

def past_window(a, window, step_size = 1):
    a = np.pad(a.astype(float), (window - 1, 0), 'constant', constant_values=(np.nan, np.nan))
    shape = a.shape[:-1] + (a.shape[-1] - window + 2 - step_size, window)
    strides = a.strides + (a.strides[-1] * step_size,)
    return  np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)

def kl(p,q):
    """
    Kullback–Leibler divergence (KL-divergence)
    """
    return -np.nansum(p[p>0]*np.log2(p[p>0]/q[p>0]))

def jenson_shannon(p,q):
    """
    Jensen–Shannon divergence
    """
    m = 0.5 * (p+q)
    return 0.5 * (kl(p,m) + kl(q,m))

def rank_array(a, ascending=True, normed=False):
    """
    Rank elements in the array.  
    ascending=> lowest=0, highest=1
    descending=> lowest=1, highest=0

    Parameters
    ----------
    a : numpy array or list
        Object to rank

    ascending : bool, default True
        Sort ascending vs. descending.

    normed : bool, default False
        False : rank is from 0 to N -1
        True : rank is from 0 to 1

    Returns
    ----------
    Ranked array.

    """
    idx = np.argsort(a)
    ranks = np.empty_like(idx)

    if ascending:
        ranks[idx] = np.arange(idx.shape[0])
    else:
        ranks[idx] = np.arange(idx.shape[0])[::-1]

    if normed:
        ranks = ranks/(ranks.shape[0]-1)
    return ranks

def convert_size(size_bytes):
    """
    Convert the bytes to human readable.
    """
    if size_bytes == 0:
       return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])

def empty_mode(a):
    """
    Mode of array when empty

    Parameters
    ----------
    a : numpy array or list
        array of values

    """
    values, counts = np.unique(a, return_counts=True)
    if values.shape[0] == 0:
        return None
    else:
        return np.random.choice(values[counts==counts.max()],1)[0]

@jit
def holder_mean(a, rho=1):
    """
    Holder mean

    Parameters
    ----------
    a : numpy array or list
        array of values

    rho : float
        holder parameter
        arithmetic mean (rho=1)
        geometric mean (rho=0)
        harmonic mean (rho=-1)
        quadratic mean (rho=2)
        max (rho-> infty)
        min (rho-> -infty)

    """
    return (a**rho).sum()**(1.0/rho) / a.shape[0]

def simpson(p):
    """
    Simpson diversity

    Parameters
    ----------
    p : numpy array or list
        array of values

    """
    un, counts = np.unique(p, return_counts=True)
    return np.square(counts/counts.sum()).sum()

def simpson_finite(p):
    """
    Simpson diversity with finite size correction

    Parameters
    ----------
    p : numpy array or list
        array of values

    """
    un, counts = np.unique(p, return_counts=True)
    N = counts.sum()
    if N > 1:
        return (counts * (counts - 1)).sum() / (N * (N - 1))
    else:
        return None

def shannon_entropy(p, base=np.e):
    """
    Shannon entropy

    Parameters
    ----------
    p : numpy array or list
        array of values

    """
    value,counts = np.unique(p, return_counts=True)
    return spstats.entropy(counts, base=base)

def value_to_int(a, sort_values='value', ascending=False, return_map=True):
    """
    Map the values of an array to integers.

    Parameters
    ----------
    a : numpy array or list
        array of values

    sort_values : str, default 'value'
        'none' : dont sort
        'value' : sort the array items based on their value
        'freq' : sort the array items based on their frequency (see ascending)

    ascending : bool, default False
        Only when sort_values == 'freq':
            False: larger counts dominate--map 0 to the most common
            True: smaller counts dominate--map 0 to the least common

    return_map: bool, default True
    """
    if sort_values == 'none':
        unique_values = pd.unique(a)
    
    elif sort_values == 'value':
        unique_values = np.sort(pd.unique(a))

    elif sort_values == 'freq':
        unique_values, freq = np.unique(a, return_counts=True)
        if ascending:
            unique_values = unique_values[np.argsort(freq)]
        else:
            unique_values = unique_values[np.argsort(freq)[::-1]]

    else:
        raise ValueError

    value_to_int_dict = {v:i for i,v in enumerate(unique_values)}
   
    if return_map:
        return [value_to_int_dict[v] for v in a], value_to_int_dict
    else:
        return [value_to_int_dict[v] for v in a]


def uniquemap_by_frequency(df, colgroupby='PublicationId', colcountby='FieldId', ascending=False):
    """
    Reduce a one-to-many mapping to a selection based on frequency of occurence in the dataframe  
    (either to the largest, most common or smallest, least common).
    
    Parameters
    ----------
    ascending : bool, default False
        False: larger counts dominate--map defaults to the most common
        True: smaller counts dominate--map defaults to the least common
    """
    countkeydict = {countid:i for i, countid in enumerate(df[colcountby].value_counts(ascending=ascending).index.values)}
    def countkey(a):
        if isinstance(a, pd.Series):
            return [countkeydict[c] for c in a]
        else:
            return countkeydict[a]
    return df.sort_values(by=[colcountby], key=countkey).drop_duplicates(subset=[colgroupby], keep='first')

@jit
def welford_mean_m2(previous_count, previous_mean, previous_m2, new_value):
    """
    Welford's algorithm for online mean and variance.
    """
    new_count = previous_count + 1
    new_mean = previous_mean + (new_value - previous_mean) / new_count
    new_m2 = previous_m2 + (new_value - previous_mean)*(new_value - new_mean)
    return (new_count, new_mean, new_m2)

@jit
def zscore_var(obs, m, var):
    return (obs-m) / np.sqrt(var)

@jit
def fast_delong(X,Y):
    """
    A fast algorithm from :cite:`Sun2014fastauc` to compute the DeLong AUC :cite:`DeLong1988auc`.
    """
    m = X.shape[0]
    n = Y.shape[0]

    if m == 0 or n == 0:
        return np.nan, np.nan

    theta = 0
    Z = np.concatenate((X,Y), axis = 0)
    Z_MR = compute_midrank(Z)
    X_MR = compute_midrank(X)
    Y_MR = compute_midrank(Y)


    auc = Z_MR[:m].sum() / m / n - float(m + 1.0) / 2.0 / n

    v01 = (Z_MR[:m] - X_MR) / n
    v10 = 1.0 - (Z_MR[m:] - Y_MR[:]) / m
    sx = np.cov(v01)
    sy = np.cov(v10)
    delongcov = sx / m + sy / n

    return auc, delongcov

@jit
def compute_midrank(x):
    """Computes midranks from :cite:`Sun2014fastauc`.

    Args:
       x - a 1D numpy array
    Returns:
       array of midranks

    """
    J = np.argsort(x)
    Z = x[J]
    N = x.shape[0]
    T = np.zeros(N, dtype=float)
    i = 0
    while i < N:
        j = i
        while j < N and Z[j] == Z[i]:
            j += 1
        T[i:j] = 0.5*(i + j - 1)
        i = j
    T2 = np.empty(N, dtype=float)
    
    T2[J] = T + 1
    return T2

def download_file_from_google_drive(file_id, destination=None, CHUNK_SIZE = 32768):
    """
    Download data files from the google Drive.

    Modified from: from https://stackoverflow.com/questions/38511444/python-download-files-from-google-drive-using-url
    """

    raise NotImplementedError("Google Drive changed their download policy and this function no longer works.  A fix will come.")

    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params = { 'id' : file_id }, stream = True)
    token = get_confirm_token(response)

    if token:
        params = { 'id' : file_id, 'confirm' : token }
        response = session.get(URL, params = params, stream = True)

    if destination is None:
        return response
    else:
        with open(destination, "wb") as f:
            for chunk in response.iter_content(CHUNK_SIZE):
                if chunk: # filter out keep-alive new chunks
                    f.write(chunk)
        return None   

def get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None


# -*- coding: utf-8 -*-
"""
.. module:: RaoStirling Interdisciplinarity
    :synopsis: Set of functions for RaoStirling Interdisciplinarity

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import pandas as pd
import numpy as np

import scipy.sparse as spsparse

from collections import defaultdict

from sklearn.metrics import pairwise_distances
from sklearn.preprocessing import normalize

from ..utils import isin_sorted, zip2dict, check4columns
from ..network import dataframe2bipartite
from .referencestrength import field_citation_vectors

def field_citation_distance(pub2ref, pub2field, pub2field_norm=True, temporal=True,citation_direction='references', 
    field_distance_metric='cosine', show_progress=False):
    """
    Calculate the field distance matrix based on references or citations.

    Parameters
    ----------
    :param pub2ref : DataFrame
        A DataFrame with the citation information for each Publication.

    :param pub2field : DataFrame
        A DataFrame with the field information for each Publication.

    :param pub2field_norm : bool, default True
        When a publication occurs in m > 1 fields, count the publication 1/m times in each field.  Normalizes the membership
        vector so it sums to 1 for each publication.

    :param temporal : bool, default False
        If True, compute the distance matrix using only publications for each year.

    :param citation_direction : str, default `references`
        `references` : the fields are defined by a publication's references.
        `citations` : the fields are defined by a publication's citations.

    :param field_distance_metric : str, default `cosine`
        The interfield distance metric.  Valid entries come from sklearn.metrics.pairwise_distances:
        ‘cosine‘, ‘euclidean’, ‘l1’, ‘l2’, etc.

    :param show_progress : bool, default False
        If True, show a progress bar tracking the calculation.

    Returns
    -------
    Distance DataFrame
        if temporal is True
            DataFrame with 4 columns: iFieldId, jFieldId, Year, and FieldDistance
        if temporal is False
            DataFrame with 3 columns: iFieldId, jFieldId, FieldDistance

    """

    # now we map citing and cited to the source and target depending on which diretion was specified by `citation_direction'
    if citation_direction == 'references':
        pub2ref_rename_dict = {'CitedPublicationId':'TargetId', 'CitingPublicationId':'SourceId'}
        year_col = 'CitingYear'
    elif citation_direction == 'citations':
        pub2ref_rename_dict = {'CitedPublicationId':'SourceId', 'CitingPublicationId':'TargetId'}
        year_col = 'CitedYear'

    required_columns = ['CitedPublicationId', 'CitingPublicationId']
    if temporal:
        required_columns.append(year_col)
    check4columns(pub2ref, required_columns)
    pub2ref = pub2ref[required_columns].dropna().copy(deep=True)

    check4columns(pub2field, ['PublicationId', 'FieldId'])
    pub2field = pub2field.copy(deep=True)

    # to leverage matrix operations we need to map fields to the rows/cols of the matrix
    field2int = {fid:i for i, fid in enumerate(np.sort(pub2field['FieldId'].unique()))}
    int2field = {i:fid for fid, i in field2int.items()}
    pub2field['FieldId'] = [field2int[fid] for fid in pub2field['FieldId'].values]
    Nfields = len(field2int)

    pub2ref.rename(columns=pub2ref_rename_dict, inplace=True)

    # the assignment of a publication to a field is 1/(number of fields) when normalized, and 1 otherwise
    if pub2field_norm:
        pub2nfields = pub2field.groupby('PublicationId')['FieldId'].nunique()
    else:
        pub2nfields = defaultdict(lambda:1)
    pub2field['PubFieldContribution'] = [1.0/pub2nfields[pid] for pid in pub2field['PublicationId'].values]

    distance = []

    # differeniate between the temporal and the static RS
    if temporal:

        for y, ydf in pub2ref.groupby(year_col):
            # merge the references to the fields for the source fields
            yfield2field_mat = field_citation_vectors(ydf, pub2field, n_fields=Nfields, count_normalize=None)

            # now compute the distance matrix for this year only
            distance_matrix = pairwise_distances(yfield2field_mat, metric=field_distance_metric)
            nnzrow, nnzcol = np.nonzero(distance_matrix)
            for isource, itarget in zip(nnzrow, nnzcol):
                if isource < itarget:
                    distance.append([int2field[isource], int2field[itarget], y, distance_matrix[isource, itarget]])

        distance = pd.DataFrame(distance, columns = ['iFieldId', 'jFieldId', year_col, 'FieldDistance'])

    else:

        field2field_mat = spsparse.coo_matrix( (Nfields, Nfields) )
        
        nref = int(pub2ref.shape[0] / 10.0**6) + 1
        for itab in range(nref):
            tabdf = pub2ref.loc[itab*10**6:(itab+1)*10**6]

            # calculate the field representation vectors
            field2field_mat += field_citation_vectors(tabdf, pub2field, n_fields=Nfields, count_normalize=None)

        # now compute the distance matrix
        distance_matrix = pairwise_distances(field2field_mat, metric=field_distance_metric)
        sources, targets = np.nonzero(distance_matrix)
        for isource, itarget in zip(sources, targets):
            if isource < itarget:
                distance.append([int2field[isource], int2field[itarget], distance_matrix[isource, itarget]])

        distance = pd.DataFrame(distance, columns = ['iFieldId', 'jFieldId', 'FieldDistance'])

    return distance
    


def raostriling_interdisciplinarity(pub2ref, pub2field, focus_pub_ids=None, pub2field_norm=True, temporal=False,
    citation_direction='references', field_distance_metric='cosine', distance_matrix=None, show_progress=False):
    """
    Calculate the RaoStirling index as a measure of a publication's interdisciplinarity.
    See :cite:`stirling20` for the definition and :cite:`gates2019naturereach` for an application.

    Parameters
    ----------
    :param pub2ref : DataFrame
        A DataFrame with the citation information for each Publication.

    :param pub2field : DataFrame
        A DataFrame with the field information for each Publication.

    :param focus_pub_ids : numpy array or list, default None
        A list of the PublicationIds to calculate interdisciplinarity.

    :param pub2field_norm : bool, default True
        When a publication occurs in m > 1 fields, count the publication 1/m times in each field.  Normalizes the membership
        vector so it sums to 1 for each publication.

    :param temporal : bool, default False
        If True, compute the distance matrix using only publications for each year.

    :param citation_direction : str, default `references`
        `references` : the fields are defined by a publication's references.
        `citations` : the fields are defined by a publication's citations.

    :param field_distance_metric : str, default `cosine`
        The interfield distance metric.  Valid entries come from sklearn.metrics.pairwise_distances:
        ‘cosine‘, ‘euclidean’, ‘l1’, ‘l2’, etc.

    :param distance_matrix : numpy array, default None
        The precomputed field distance matrix.

    :param show_progress : bool, default False
        If True, show a progress bar tracking the calculation.

    Returns
    -------
    DataFrame
        DataFrame with 2 columns: 'PublicationId', 'RaoStirling'

    """

    # now we map citing and cited to the source and target depending on which diretion was specified by `citation_direction'
    if citation_direction == 'references':
        pub2ref_rename_dict = {'CitedPublicationId':'TargetId', 'CitingPublicationId':'SourceId'}
        year_col = 'CitingYear'
    elif citation_direction == 'citations':
        pub2ref_rename_dict = {'CitedPublicationId':'SourceId', 'CitingPublicationId':'TargetId'}
        year_col = 'CitedYear'

    required_columns = ['CitedPublicationId', 'CitingPublicationId']
    if temporal:
        required_columns.append(year_col)
    check4columns(pub2ref, required_columns)
    pub2ref = pub2ref[required_columns].dropna().copy(deep=True)

    check4columns(pub2field, ['PublicationId', 'FieldId'])
    pub2field = pub2field.copy(deep=True)

    # check that the precomputed distance matrix is the correct size
    if distance_matrix is None:
        distance_matrix = field_citation_distance(pub2ref, pub2field, pub2field_norm, temporal, 
            citation_direction, field_distance_metric, show_progress)

    field2int = {fid:i for i, fid in enumerate(np.sort(pub2field['FieldId'].unique()))}
    pub2field['FieldId'] = [field2int[fid] for fid in pub2field['FieldId'].values]
    Nfields = len(field2int)

    pub2ref.rename(columns=pub2ref_rename_dict, inplace=True)

    if not focus_pub_ids is None:
        pub2ref = pub2ref.loc[isin_sorted(pub2ref['SourceId'].values, focus_pub_ids)]

    if temporal:
        years = np.sort(pub2ref[year_col].unique())
        year2int = {y:i for i, y in enumerate(years)}
        Nyears = years.shape[0]

    if type(distance_matrix) == pd.DataFrame and temporal:
        check4columns(distance_matrix, ['iFieldId', 'jFieldId', year_col, 'FieldDistance'])

        distance_matrix = distance_matrix.loc[isin_sorted(distance_matrix[year_col].values, years)].copy(deep=True)

        distance_matrix['iFieldId'] = [field2int.get(fid, None) for fid in distance_matrix['iFieldId'].values]
        distance_matrix['jFieldId'] = [field2int.get(fid, None) for fid in distance_matrix['jFieldId'].values]
        distance_matrix.dropna(inplace=True)

        tdm = np.zeros((Nyears, Nfields, Nfields))
        for y in years:
            tdm[year2int[y]] = dataframe2bipartite(df=distance_matrix[distance_matrix[year_col] == y], rowname='iFieldId', colname='jFieldId',
                shape=(Nfields, Nfields), weightname='FieldDistance').todense()

            tdm[year2int[y]] = tdm[year2int[y]] + tdm[year2int[y]].T

        distance_matrix = tdm


    elif type(distance_matrix) == pd.DataFrame and not temporal:
        check4columns(distance_matrix, ['iFieldId', 'jFieldId', 'FieldDistance'])
        distance_matrix = distance_matrix.copy(deep=True)
        distance_matrix['iFieldId'] = [field2int.get(fid, None) for fid in distance_matrix['iFieldId'].values]
        distance_matrix['jFieldId'] = [field2int.get(fid, None) for fid in distance_matrix['jFieldId'].values]
        distance_matrix.dropna(inplace=True)
        distance_matrix = dataframe2bipartite(df=distance_matrix, rowname='iFieldId', colname='jFieldId',
                shape=(Nfields, Nfields), weightname='FieldDistance').todense()

        distance_matrix = distance_matrix + distance_matrix.T

    elif (type(distance_matrix) == np.array or type(distance_matrix) == np.matrix):
        if not temporal and distance_matrix.shape != (Nfields, Nfields):
            raise pySciSciMetricError('The precomputed_distance_matrix is of the wrong size to compute the RaoStirling interdisciplinarity for the publications passed.')
        elif temporal and distance_matrix.shape != (Nyears, Nfields, Nfields):
            raise pySciSciMetricError('The precomputed_distance_matrix is of the wrong size to compute the RaoStirling interdisciplinarity for the publications and years passed.')

    # the assignment of a publication to a field is 1/(number of fields) when normalized, and 1 otherwise
    if pub2field_norm:
        pub2nfields = pub2field.groupby('PublicationId')['FieldId'].nunique()
    else:
        pub2nfields = defaultdict(lambda:1)
    pub2field['PubFieldContribution'] = [1.0/pub2nfields[pid] for pid in pub2field['PublicationId'].values]

    # merge the references to the fields for the target fields
    pub2ref = pub2ref.merge(pub2field, how='left', left_on='TargetId', right_on='PublicationId').rename(
        columns={'FieldId':'TargetFieldId', 'PubFieldContribution':'TargetPubFieldContribution'})
    del pub2ref['PublicationId']

    pub2ref.dropna(inplace=True)

    # Now we start on the RaoStiring calculation
    if temporal:

        rsdf = []
        for y, ydf in pub2ref.groupby(year_col):
            
            # for each year, we need to map individual publications to the rows of our matrix
            ypub2int = {pid:i for i, pid in enumerate(np.sort(ydf['SourceId'].unique()))}
            yint2pub = {i:pid for pid, i in ypub2int.items()}
            ydf['SourceId'] = [ypub2int[fid] for fid in ydf['SourceId'].values]
            yNpubs = len(ypub2int)

            # calculate the publication representation vectors over fields
            ypub2field_mat = dataframe2bipartite(df=ydf, rowname='SourceId', colname='TargetFieldId',
                shape=(yNpubs, Nfields), weightname='TargetPubFieldContribution').tocsr()

            # make sure the publication 2 field vector is normalized
            ypub2field_mat = normalize(ypub2field_mat, norm='l1', axis=1)

            # finally, we calculate the matrix representation of the RS measure
            yrsdf = pd.DataFrame()
            yrsdf['PublicationId'] = [yint2pub[i] for i in np.sort(ydf['SourceId'].unique())]
            yrsdf['CitingYear'] = y
            yrsdf['RaoStirling'] =  0.5 * np.squeeze(np.asarray(ypub2field_mat.dot(spsparse.csr_matrix(distance_matrix[year2int[y]])).multiply(ypub2field_mat).sum(axis=1)))
            
            rsdf.append(yrsdf)

        rsdf = pd.concat(rsdf)

        return rsdf

    else:

        # first map individual publications to the rows of our matrix
        pub2int = {pid:i for i, pid in enumerate(np.sort(pub2ref['SourceId'].unique()))}
        int2pub = {i:pid for pid, i in pub2int.items()}
        pub2ref['SourceId'] = [pub2int[pid] for pid in pub2ref['SourceId'].values]
        pub2ref[['SourceId', 'TargetFieldId']] = pub2ref[['SourceId', 'TargetFieldId']].astype(int)
        Npubs = len(pub2int)

        # calculate the publication representation vectors over fields
        pub2field_mat = dataframe2bipartite(df=pub2ref, rowname='SourceId', colname='TargetFieldId',
                shape=(Npubs, Nfields), weightname='TargetPubFieldContribution').tocsr()

        # make sure the publication 2 field vector is normalized
        pub2field_mat = normalize(pub2field_mat, norm='l1', axis=1)

        distance_matrix = spsparse.csr_matrix(distance_matrix)

        # finally, we calculate the matrix representation of the RS measure
        rsdf = pd.DataFrame()
        rsdf['RaoStirling'] = 0.5 * np.squeeze(np.asarray( spsparse.csr_matrix.multiply(pub2field_mat.dot(distance_matrix), pub2field_mat).sum(axis=1)))
        rsdf['PublicationId'] = [int2pub[i] for i in np.sort(pub2ref['SourceId'].unique())]
        
        return rsdf




# -*- coding: utf-8 -*-
"""
.. module:: novelty
    :synopsis: Set of functions to quantify the novelty and conventionality

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import pandas as pd
import numpy as np

from itertools import combinations

from pyscisci.utils import isin_sorted, groupby_count, value_to_int, welford_mean_m2, zscore_var, check4columns
from pyscisci.sparsenetworkutils import dataframe2bipartite, project_bipartite_mat

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
import sys
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

### Novelty

def novelty_conventionality(pub, pub2ref, focus_pub_ids=None, n_samples = 10, show_progress=False):

    """
    This function calculates the novelty and conventionality for publications as proposed in :cite:`Uzzi2013atypical`.
    
    ToDo: path2randomnetwork, save randomizations


    Parameters
    ----------
    pub : DataFrame
        A DataFrame with Year and Journal information for each Publication.

    pub2ref : DataFrame
        A DataFrame with the reference information for each Publication.

    focus_pub_ids : list or numpy array, default None
        A list of PublicationIds for which to compute the novelty score.

    n_samples : int, default 10
        The number of randomized networks in the ensemble.

    show_progress : bool, default False
        If True, show a progress bar tracking the calculation.

    Returns
    -------
    DataFrame
        The original dataframe with a new column for rank: colrankby+"Rank"

    """

    desc=''
    if isinstance(show_progress, str):
        desc = show_progress

    #first subset only to the data columns we need
    required_pub_columns = ['PublicationId', 'Year', 'JournalId']
    check4columns(pub, required_pub_columns)
    pub = pub[required_pub_columns].dropna()

    required_pub2ref_columns = ['CitingPublicationId', 'CitedPublicationId']
    check4columns(pub2ref, required_pub2ref_columns)
    pub2ref = pub2ref[required_pub2ref_columns]

    # now map the journal values to integers so we can use them for matrix indices
    pub['JournalInt'], journal2int = value_to_int(pub['JournalId'].values, sort_values='value', return_map=True)
    Njournals = len(journal2int)

    # now map the publication values to integers so we can use them for matrix indices
    pub['PublicationInt'], pub2int = value_to_int(pub['PublicationId'].values, sort_values='value', return_map=True)
    Npubs = pub['PublicationInt'].nunique()
    int2pub = {i:pid for pid, i in pub2int.items()}

    pub2ref = pub2ref.merge(pub[['PublicationId', 'PublicationInt', 'Year']], how='left', left_on='CitingPublicationId', right_on='PublicationId')
    del pub2ref['PublicationId']
    pub2ref.rename(columns={'PublicationInt':'CitingPublicationInt', 'Year':'CitingYear'}, inplace=True)

    pub2ref = pub2ref.merge(pub[['PublicationId', 'Year', 'JournalInt']], how='left', left_on='CitedPublicationId', right_on='PublicationId')
    del pub2ref['PublicationId']
    del pub2ref['CitedPublicationId']
    pub2ref.rename(columns={'JournalInt':'CitedJournalInt', 'Year':'CitedYear'}, inplace=True)

    pub2ref = pub2ref.dropna().reset_index(drop=True)

    # book keeping for our focus publications
    if isinstance(focus_pub_ids, list) or isinstance(focus_pub_ids, np.ndarray):
        focus_pub_ids = np.sort(focus_pub_ids)
    elif isinstance(focus_pub_ids, int) or isinstance(focus_pub_ids, str):
        focus_pub_ids = np.sort([focus_pub_ids])

    # book keeping for the focus years
    if focus_pub_ids is None:
        years = np.sort(pub2ref['CitingYear'].unique())
    else:
        years = np.sort(pub2ref[isin_sorted(pub2ref['CitingPublicationId'].values, focus_pub_ids)]['CitingYear'].unique())

    # The scores are found for all publications in one year, so we can treat each year independently
    novelty_score = []
    for citing_year in tqdm(years, desc="Novelty_Conventoinality", leave=True, disable=not show_progress):

        # subset to the focus year
        ypub2ref = pub2ref.loc[pub2ref['CitingYear'] == citing_year].reset_index(drop=True)
        
        # find the observed journal-pair frequencies
        observed_journal_bipartite = dataframe2bipartite(ypub2ref, 'CitedJournalInt', 'CitingPublicationInt', (Njournals, Npubs) )
        observed_journal_adj = project_bipartite_mat(observed_journal_bipartite, project_to = 'row').todense()

        # copy over the table
        database_table = ypub2ref.copy()

        random_means = np.zeros((Njournals, Njournals))
        random_m2s = np.zeros((Njournals, Njournals))

        # run this loop for each sample
        for isample in range(n_samples):

            # the MCMC step discussed in the Supplemental for :cite:`Uzzi2013atypical` is equivalent to a hard-shuffle of the journal column 
            database_table['CitedJournalInt'] = database_table.groupby(['CitedYear'], sort=False)['CitedJournalInt'].transform(np.random.permutation)
            
            # now find the random journal pair frequencies
            random_journal_bipartite = dataframe2bipartite(database_table, rowname='CitedJournalInt', colname='CitingPublicationInt', shape=(Njournals, Npubs))
            random_journal_adj = project_bipartite_mat(random_journal_bipartite, project_to = 'row').todense()

            # update our running means and m2 (mean squared errors) using Welford's algorithm
            # this means we dont have to keep the random networks in memory and instead can just update our statistics for each sample!
            _, random_means, random_m2s = welford_mean_m2(isample, random_means, random_m2s, random_journal_adj)
        
        # find the journal pair z-scores   
        # a lot of the entries will still be zero, so lets hide the warnings when we calculate the z-score
        with np.errstate(divide='ignore',invalid='ignore'):
            journal_zscores = zscore_var(observed_journal_adj, random_means, random_m2s/n_samples)

        # now we need to map over the references to each publication
        def _ref_distribution(refentry):
            if refentry.shape[0] > 1:
                # find the zscore for each pair of references
                ref_pair_zscores = [journal_zscores[i,j] for i,j in combinations(refentry.values, 2)]

                # and return the 10th percentile (novelty) and median (conventionality)
                return pd.Series([np.percentile(ref_pair_zscores, 10), np.median(ref_pair_zscores)], 
                                 index=['NoveltyScore', 'ConventionalityScore'])
            else:
                return pd.Series([None, None], 
                                 index=['NoveltyScore', 'ConventionalityScore'])

        with np.errstate(divide='ignore',invalid='ignore'):
            ynovelty_score = ypub2ref.groupby('CitingPublicationInt')['CitedJournalInt'].apply(_ref_distribution)

        # finally do some reordering to make the returned dataframe nice
        ynovelty_score = ynovelty_score.unstack(1).reset_index().rename(columns={'CitingPublicationInt':'PublicationId'})
        ynovelty_score['PublicationId'] = [int2pub.get(i, None) for i in ynovelty_score['PublicationId'].values]
        if not focus_pub_ids is None:
            ynovelty_score = ynovelty_score.loc[isin_sorted(ynovelty_score['PublicationId'].values, focus_pub_ids)]
        
        novelty_score.append(ynovelty_score)

    novelty_score = pd.concat(novelty_score)

    return novelty_score



# -*- coding: utf-8 -*-
"""
.. module:: publicationmetrics
    :synopsis: Set of functions for the bibliometric analysis of publications

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import sys

import pandas as pd
import numpy as np

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.utils import rank_array, check4columns

from pyscisci.methods.raostirling import *
from pyscisci.methods.diversity import *
from pyscisci.methods.creditshare import *
from pyscisci.methods.disruption import *
from pyscisci.methods.longtermimpact import *
from pyscisci.methods.sleepingbeauty import *
from pyscisci.methods.pivotscore import *
from pyscisci.methods.novelty import *


def citation_rank(df, colgroupby='Year', colrankby='C10', ascending=True, normed=False, show_progress=False):
    """
    Rank publications by the number of citations (smallest) to N -1 (largest)

    Parameters
    ----------
    df : DataFrame
        A DataFrame with the citation information for each Publication.

    colgroupby : str, list
        The DataFrame column(s) to subset by.

    colrankby : str
        The DataFrame column to rank by.

    ascending : bool, default True
        Sort ascending vs. descending.

    normed : bool, default False
        - False : rank is from 0 to N -1
        - True : rank is from 0 to 1

    show_progress : bool, default False
        If True, show a progress bar tracking the calculation.

    Returns
    -------
    DataFrame
        The original dataframe with a new column for rank: colrankby+"Rank"

    """
    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc='Citation Rank', disable= not show_progress)

    df[str(colrankby)+"Rank"] = df.groupby(colgroupby)[colrankby].progress_transform(lambda x: rank_array(x, ascending, normed))
    return df

def publication_beauty(pub2ref , colgroupby = 'CitedPublicationId', colcountby = 'CitingPublicationId', show_progress=False):
    """
    Calculate the sleeping beauty and awakening time for each cited publication.  See :cite:`Sinatra2016qfactor` for the derivation.

    The algorithmic implementation can be found in :py:func:`metrics.qfactor`.

    Parameters
    ----------
    pub2ref  : DataFrame, default None, Optional
        A DataFrame with the temporal citing information information.

    colgroupby : str, default 'CitedPublicationId', Optional
        The DataFrame column with Author Ids.  If None then the database 'CitedPublicationId' is used.

    colcountby : str, default 'CitingPublicationId', Optional
        The DataFrame column with Citation counts for each publication.  If None then the database 'CitingPublicationId' is used.

    Returns
    -------
    DataFrame
        Trajectory DataFrame with 2 columns: 'AuthorId', 'Hindex'

    """

    check4columns(pub2ref , ['CitedPublicationId', 'CitingPublicationId', 'CitingYear'])

    tqdm.pandas(desc='Beauty', disable= not show_progress)

    df = groupby_count(pub2ref , colgroupby = ['CitedPublicationId', 'CitingYear'], colcountby = 'CitingPublicationId', count_unique = True)

    newname_dict = zip2dict([str(colcountby), '0', '1'], [str(colgroupby)+'Beauty']*2 + ['Awakening'])
    return df.groupby(colgroupby)[colcountby + 'Count'].progress_transform(beauty_coefficient).rename(columns=newname_dict)



# -*- coding: utf-8 -*-
"""
.. module:: journalcitation
    :synopsis: Set of functions for typical journal bibliometric citation analysis

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import sys

import pandas as pd
import numpy as np

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.utils import zip2dict, groupby_count
from pyscisci.methods.hindex import compute_hindex

def journal_productivity(pub2journal, colgroupby = 'JournalId', colcountby = 'PublicationId', show_progress=False):
    """
    Calculate the total number of publications for each journal.

    Parameters
    ----------
    :param pub2journal : DataFrame, default None, Optional
        A DataFrame with the author2publication information.

    :param colgroupby : str, default 'JournalId', Optional
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    :param colcountby : str, default 'PublicationId', Optional
        The DataFrame column with Publication Ids.  If None then the database 'PublicationId' is used.


    Returns
    -------
    DataFrame
        Productivity DataFrame with 2 columns: 'AuthorId', 'Productivity'

    """

    # we can use show_progress to pass a label for the progress bar
    if show_progress:
        show_progress='Journal Productivity'

    newname_dict = zip2dict([str(colcountby)+'Count', '0'], ['Productivity']*2)
    return groupby_count(pub2journal, colgroupby, colcountby, count_unique=True, show_progress=show_progress).rename(columns=newname_dict)

def journal_yearly_productivity(pub2journal, colgroupby = 'JournalId', datecol = 'Year', colcountby = 'PublicationId', show_progress=False):
    """
    Calculate the number of publications for each author in each year.

    Parameters
    ----------
    :param pub2journal : DataFrame, default None, Optional
        A DataFrame with the author2publication information. 

    :param colgroupby : str, default 'AuthorId', Optional
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    :param datecol : str, default 'Year', Optional
        The DataFrame column with Year information.  If None then the database 'Year' is used.

    :param colcountby : str, default 'PublicationId', Optional
        The DataFrame column with Publication Ids.  If None then the database 'PublicationId' is used.

    Returns
    -------
    DataFrame
        Productivity DataFrame with 3 columns: 'AuthorId', 'Year', 'YearlyProductivity'

    """
    
    # we can use show_progress to pass a label for the progress bar
    if show_progress:
        show_progress='Journal Yearly Productivity'

    newname_dict = zip2dict([str(colcountby)+'Count', '0'], ['YearlyProductivity']*2)
    return groupby_count(pub2journal, [colgroupby, datecol], colcountby, count_unique=True, show_progress=show_progress).rename(columns=newname_dict)

def journal_hindex(pub2journal, impact=None, colgroupby = 'JournalId', colcountby = 'Ctotal', show_progress=False):
    """
    Calculate the author yearly productivity trajectory.  See :cite:`hirsch2005index` for the derivation.

    The algorithmic implementation can be found in :py:func:`citationanalysis.compute_hindex`.

    Parameters
    ----------
    :param pub2journal : DataFrame
        A DataFrame with the publication and journal information.

    :param impact : DataFrame, default None, Optional
        A DataFrame with the publication citation counts precalculated.  If None, then it is assumed that the citation counts are already in pub2journal.

    :param colgroupby : str, default 'JournalId', Optional
        The DataFrame column with Author Ids.  If None then the database 'JournalId' is used.

    :param colcountby : str, default 'Ctotal', Optional
        The DataFrame column with Citation counts for each publication.  If None then the database 'Ctotal' is used.

    :param show_progress : bool, default False
        The DataFrame column with Citation counts for each publication.  If None then the database 'Ctotal' is used.
    
    :param show_progress: bool, default False
            Show progress of the calculation.

    Returns
    -------
    DataFrame
        Trajectory DataFrame with 2 columns: 'JournalId', 'Hindex'

    """
    if not impact is None:
        pub2journal = pub2journal.merge(impact[[colgroupby, colcountby]], on='PublicationId', how='left')

    if show_progress: print("Computing Journal H-index.")
    return compute_hindex(pub2journal, colgroupby = colgroupby, colcountby = colcountby, show_progress=show_progress)

def journal_impactfactor(pub, pub2ref, pub2year=None, citation_window=5, colgroupby = 'JournalId', show_progress=False):
    """
    Calculate the impact factor for a journal.

    Parameters
    ----------
    :param pub : DataFrame
        A DataFrame with the publication, journal, and year information.

    :param pub2ref : DataFrame
        A DataFrame with the author2publication information.  If None then the database 'author2pub' is used.

    :param pub2year : dict, defualt None, Optional
        A dictionary mapping 'PublicationIds' to the publication year.  If None then the 'CitingYear' is assumed to be a column of pub2ref.

    :param colgroupby : str, default 'JournalId', Optional
        The DataFrame column with Author Ids.  If None then the database 'JournalId' is used.

    :param colcountby : str, default 'Ctotal', Optional
        The DataFrame column with Citation counts for each publication.  If None then the database 'Ctotal' is used.
    
    :param show_progress: bool, default False
            Show progress of the calculation.

    Returns
    -------
    DataFrame
        Trajectory DataFrame with 2 columns: 'JournalId', 'ImpactFactor{y}' where y is the citation_window size

    """




# -*- coding: utf-8 -*-
"""
.. module:: productivitytrajectory
    :synopsis: Calculate the productivity trajectory.

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import sys

import pandas as pd
import numpy as np

import scipy.optimize as spopt

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.utils import zip2dict


### Productivity Trajectory

def piecewise_linear(x, x_break, b, m1, m2):
    """
    A piece wise linear function:
    x <= x_break: y = m1*x + b
    x > x_break : y = m1*x_break + b + m2*(x - x_break)
    """
    return np.piecewise(x, [x <= x_break], [lambda x:m1*x + b-m1*x_break, lambda x:m2*x + b-m2*x_break])

def fit_piecewise_linear(xvalues, yvalues):
    m0 = (yvalues.max()-yvalues.min())/(xvalues.max() - xvalues.min())
    p0 = [np.median(xvalues), yvalues.mean(), m0, m0]
    p , e = spopt.curve_fit(piecewise_linear, xvalues, yvalues, p0=p0)
    return pd.Series(p)


def _fit_piecewise_lineardf(author, args):
    return fit_piecewise_linear(author[args[0]].values, author[args[1]].values)

def yearly_productivity_traj(df, colgroupby = 'AuthorId', colx='Year',coly='YearlyProductivity'):
    """
    Calculate the piecewise linear yearly productivity trajectory original studied in :cite:`way2017misleading`.

    """

    newname_dict = zip2dict(list(range(4)), ['t_break', 'b', 'm1', 'm2' ]) #[str(i) for i in range(4)]
    return df.groupby(colgroupby, sort=False).apply(_fit_piecewise_lineardf, args=(colx,coly) ).reset_index().rename(columns = newname_dict)




# -*- coding: utf-8 -*-
"""
.. module:: reference strength
    :synopsis: Set of functions for finding field reference strength and share.

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import pandas as pd
import numpy as np

from collections import defaultdict

import scipy.sparse as spsparse

from sklearn.preprocessing import normalize

from ..utils import isin_sorted, zip2dict, check4columns
from ..network import dataframe2bipartite


def field_citation_vectors(source2target, pub2field, count_normalize=None, n_fields = None, citation_direction='references'):
    """
    Calculate the citation vector of reference/citation counts from a field to another field.

    Parameters
    ----------
    :param pub2reffield : DataFrame
        A DataFrame with the citation relationships and fields for each Publication.

    count_normalize : int, default None
        If None: use the raw reference/citation counts.
        If 0: Normalize by the Target Counts
        If 1: Normalize by the Source Counts

    :param citation_direction : str, default `references`
        `references` : the fields are defined by a publication's references.
        `citations` : the fields are defined by a publication's citations.

    Returns
    -------
    Distance DataFrame
        DataFrame with 3 columns: SourceFieldId, TargetFieldId, Count

    """

    if n_fields is None:
        Nfields = int(pub2field['FieldId'].max()) + 1
    else:
        Nfields = n_fields

    source2target = source2target.merge(pub2field[['PublicationId', 'FieldId', 'PubFieldContribution']], how='left', left_on='SourceId', right_on='PublicationId').rename(
            columns={'FieldId':'SourceFieldId', 'PubFieldContribution':'SourcePubFieldContribution'})
    del source2target['PublicationId']

    source2target = source2target.merge(pub2field[['PublicationId', 'FieldId', 'PubFieldContribution']], how='left', left_on='TargetId', right_on='PublicationId').rename(
    columns={'FieldId':'TargetFieldId', 'PubFieldContribution':'TargetPubFieldContribution'})
    del source2target['PublicationId']

    # drop any citation relationships for which we dont have field information
    source2target.dropna(inplace=True)

    # we need to use integer ids to map to the matrix
    source2target[['SourceFieldId', 'TargetFieldId']] = source2target[['SourceFieldId', 'TargetFieldId']].astype(int)

    # in the field2field distance matrix, the weighted contribution from a source publication in multiple fields
    # is the product of the source and target contributions
    source2target['SourcePubFieldContribution'] = source2target['SourcePubFieldContribution'] * source2target['TargetPubFieldContribution']

    # calculate the field representation vectors for this year only
    field2field_mat = dataframe2bipartite(df=source2target, rowname='SourceFieldId', colname='TargetFieldId',
            shape=(Nfields, Nfields), weightname='SourcePubFieldContribution')

    # now normalize the vectors by the source/target counts
    if not count_normalize is None:
        field2field_mat = normalize(field2field_mat, norm='l1', axis=count_normalize)

    return field2field_mat

def field_citation_share(pub2ref, pub2field, count_normalize=None, pub2field_norm=True, temporal=True, 
    citation_direction='references', blocksize = 10**6, show_progress=False):
    """
    Calculate the field citation share based on references or citations.

    Parameters
    ----------
    :param pub2ref : DataFrame
        A DataFrame with the citation information for each Publication.

    :param pub2field : DataFrame
        A DataFrame with the field information for each Publication.

    count_normalize : int, default None
        If None: use the raw reference/citation counts.
        If 0: Normalize by the Target Counts
        If 1: Normalize by the Source Counts

    :param pub2field_norm : bool, default True
        When a publication occurs in m > 1 fields, count the publication 1/m times in each field.  Normalizes the membership
        vector so it sums to 1 for each publication.

    :param temporal : bool, default False
        If True, compute the Share matrix using only publications for each year.

    :param citation_direction : str, default `references`
        `references` : the fields are defined by a publication's references.
        `citations` : the fields are defined by a publication's citations.

    :param show_progress : bool, default False
        If True, show a progress bar tracking the calculation.

    Returns
    -------
    CitationShare DataFrame
        if temporal is True
            DataFrame with 4 columns: iFieldId, jFieldId, Year, and Share
        if temporal is False
            DataFrame with 3 columns: iFieldId, jFieldId, Share

    """

    # now we map citing and cited to the source and target depending on which diretion was specified by `citation_direction'
    if citation_direction == 'references':
        pub2ref_rename_dict = {'CitedPublicationId':'TargetId', 'CitingPublicationId':'SourceId'}
        year_col = 'CitingYear'
    elif citation_direction == 'citations':
        pub2ref_rename_dict = {'CitedPublicationId':'SourceId', 'CitingPublicationId':'TargetId'}
        year_col = 'CitedYear'

    required_columns = ['CitedPublicationId', 'CitingPublicationId']
    if temporal:
        required_columns.append(year_col)
    check4columns(pub2ref, required_columns)
    pub2ref = pub2ref[required_columns].dropna().copy(deep=True)

    check4columns(pub2field, ['PublicationId', 'FieldId'])
    pub2field = pub2field.copy(deep=True)

    # to leverage matrix operations we need to map fields to the rows/cols of the matrix
    field2int = {fid:i for i, fid in enumerate(np.sort(pub2field['FieldId'].unique()))}
    int2field = {i:fid for fid, i in field2int.items()}
    pub2field['FieldId'] = [field2int[fid] for fid in pub2field['FieldId'].values]
    Nfields = len(field2int)

    pub2ref.rename(columns=pub2ref_rename_dict, inplace=True)

    # the assignment of a publication to a field is 1/(number of fields) when normalized, and 1 otherwise
    if pub2field_norm:
        pub2nfields = pub2field.groupby('PublicationId')['FieldId'].nunique()
    else:
        pub2nfields = defaultdict(lambda:1)
    pub2field['PubFieldContribution'] = [1.0/pub2nfields[pid] for pid in pub2field['PublicationId'].values]

    citation_share = []

    # differeniate between the temporal and the static RS
    if temporal:

        for y, ydf in pub2ref.groupby(year_col):
            
            yfield2field_mat = field_citation_vectors(ydf, pub2field, n_fields=Nfields, count_normalize=count_normalize)

            nnzrow, nnzcol = np.nonzero(yfield2field_mat)
            for isource, itarget in zip(nnzrow, nnzcol): 
                if isource < itarget:
                    citation_share.append([int2field[isource], int2field[itarget], y, yfield2field_mat[isource, itarget]])

        citation_share = pd.DataFrame(citation_share, columns = ['SourceFieldId', 'TargetFieldId', year_col, 'Share'])

    else:

        field2field_mat = spsparse.coo_matrix( (Nfields, Nfields) )
        
        nref = int(pub2ref.shape[0] / float(blocksize)) + 1
        for itab in range(nref):
            tabdf = pub2ref.loc[itab*int(blocksize):(itab+1)*int(blocksize)]
            
            field2field_mat += field_citation_vectors(tabdf, pub2field, n_fields=Nfields, count_normalize=None)

        # now normalize the vectors by the source/target counts
        if not count_normalize is None:
            field2field_mat = normalize(field2field_mat, norm='l1', axis=count_normalize)

        # now compute the citation_share matrix
        sources, targets = np.nonzero(field2field_mat)
        for isource, itarget in zip(sources, targets):
            citation_share.append([int2field[isource], int2field[itarget], field2field_mat[isource, itarget]])

        citation_share = pd.DataFrame(citation_share, columns = ['SourceFieldId', 'TargetFieldId', 'Share'])

    return citation_share
    
def field_citation_strength(pub2ref, pub2field, count_normalize=None, pub2field_norm=True, temporal=True, 
    citation_direction='references', baseline_year = 1950, show_progress=False):
    """
    Calculate the field citation strength based on references or citations.

    Parameters
    ----------
    :param pub2ref : DataFrame
        A DataFrame with the citation information for each Publication.

    :param pub2field : DataFrame
        A DataFrame with the field information for each Publication.

    count_normalize : int, default None
        If None: use the raw reference/citation counts.
        If 0: Normalize by the Target Counts
        If 1: Normalize by the Source Counts

    :param pub2field_norm : bool, default True
        When a publication occurs in m > 1 fields, count the publication 1/m times in each field.  Normalizes the membership
        vector so it sums to 1 for each publication.

    :param temporal : bool, default False
        If True, compute the Strength matrix using only publications for each year.

    :param citation_direction : str, default `references`
        `references` : the fields are defined by a publication's references.
        `citations` : the fields are defined by a publication's citations.

    :param show_progress : bool, default False
        If True, show a progress bar tracking the calculation.

    Returns
    -------
    Strength DataFrame
        if temporal is True
            DataFrame with 4 columns: iFieldId, jFieldId, Year, and Strength
        if temporal is False
            DataFrame with 3 columns: iFieldId, jFieldId, Strength

    """

    if citation_direction == 'references':
        strength_denom = 'TargetFieldId'
        year_col = 'CitingYear'
    elif citation_direction == 'citations':
        strength_denom = 'SourceFieldId'
        year_col = 'CitedYear'

    citation_strength = field_citation_share(pub2ref, pub2field, count_normalize=count_normalize, pub2field_norm=pub2field_norm, 
        temporal=temporal, citation_direction=citation_direction, show_progress=show_progress).rename(columns={'Share':'Strength'})

    pub2field = pub2field[pub2field['Year'] >= baseline_year]

    if temporal:
        for y in np.sort(citation_strength[year_col].unique()):
            yfield_size = pub2field[pub2field['Year'] <= y]['FieldId'].value_counts()

            yfield_size = yfield_size / yfield_size.sum()

            year_rows = citation_strength[year_col] == y
            citation_strength.loc[year_rows, 'Strength'] = [s/yfield_size[normid] for normid, s in citation_strength[year_rows][[strength_denom, 'Strength']].values]

    else:

        field_size = pub2field['FieldId'].value_counts()
        field_size = field_size / field_size.sum()

        citation_strength['Strength'] = [s/field_size[normid] for normid, s in citation_strength[[strength_denom, 'Strength']].values]

    return citation_strength





import os
import sys
import pathlib
import pandas as pd
import numpy as np

import dask.dataframe as dd

from unidecode import unidecode
import html

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


from pyscisci.utils import isin_sorted

def load_int(v):
    try:
        return int(v)
    except (ValueError, TypeError):
        return None

def load_float(v):
    try:
        return float(v)
    except (ValueError, TypeError):
        return None

def load_bool(v):
    try:
        if isinstance(v, str):
            if v == 'false' or v=='False' or v=='FALSE':
                return False
            elif v == 'true' or v=='True' or v=='TRUE':
                return True
            else:
                return None
        elif isinstance(v, int):
            if v == 0 or v==False:
                return False
            elif v==1 or v==True:
                return True
            else:
                return None
        else:
            return None
    except (ValueError, TypeError):
        return None

def load_html_str(s):
    if s is None:
        return ''
    else:
        return unidecode(html.unescape(s)).strip()

def load_xml_text(root_element, default=''):
    if root_element is None or len(root_element) == 0:
        try:
            return root_element.text
        except:
            return default
    else:
        return root_element[0].text


def load_preprocessed_data(dataname, path2database, database_extension = 'hdf', columns = None, filter_dict=None, duplicate_subset=None,
    duplicate_keep='last', dropna=None, keep_source_file=False, prefunc2apply=None, postfunc2apply=None, use_dask=False, show_progress=False):
    """
        Load the preprocessed DataFrame from a preprocessed directory.

        Parameters
        ----------
        :param dataname : str
            The type of preprocessed data to load.

        :param path2database : str
            The path to the database directory.

        :param columns : list, default None
            Load only this subset of columns

        :param filter_dict : dict, default None
            Dictionary of format {"ColumnName":"ListofValues"} where "ColumnName" is a data column
            and "ListofValues" is a sorted list of valid values.  A DataFrame only containing rows that appear in
            "ListofValues" will be returned.

        :param duplicate_subset : list, default None
            Drop any duplicate entries as specified by this subset of columns

        :param duplicate_keep : str, default 'last', Optional
            If duplicates are being dropped, keep the 'first' or 'last'
            (see `pandas.DataFram.drop_duplicates <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html>`_)

        :param dropna : list, default None, Optional
            Drop any NaN entries as specified by this subset of columns

        :param keep_source_file : bool, default False
            Keep track of the source file the data was loaded from.

        :param prefunc2apply : callable, default None
            A function to apply to each of the sub-DataFrames as they are loaded before filtering.

        :param postfunc2apply : callable, default None
            A function to apply to each of the sub-DataFrames as they are loaded after filtering.

        Returns
        -------
        DataFrame
            dataname DataFrame.

    """

    path2files = os.path.join(path2database, dataname)
    if not os.path.exists(path2files):
        # TODO: make a real warning
        raise NotImplementedError("First preprocess the raw data.")
        return []

    if isinstance(columns, str):
        columns = [columns]

    if isinstance(dropna, str):
        dropna = [dropna]

    if isinstance(duplicate_subset, str):
        duplicate_subset = [duplicate_subset]

    if isinstance(filter_dict, dict):
        filter_dict = {isinkey:np.sort(isinlist) for isinkey, isinlist in filter_dict.items()}


    FileNumbers = sorted([int(fname.replace(dataname, '').split('.')[0]) for fname in os.listdir(path2files) if dataname in fname and database_extension in fname])

    desc=''
    if isinstance(show_progress, str):
        desc = show_progress

    if database_extension == 'hdf' and not use_dask:
        def read_file(fname):
            return pd.read_hdf(fname, key=dataname)
    elif database_extension == 'hdf' and use_dask:
        raise NotImplementedError('Dask is not compatible with hdf dataframes.  Please reprocess using csv or csv.gz (recommended).')
    elif database_extension == 'csv' and not use_dask:
        def read_file(fname):
            return pd.read_csv(fname)
    elif database_extension == 'csv' and use_dask:
        def read_file(fname):
            return dd.read_csv(fname)
    elif database_extension == 'csv.gz' and not use_dask:
        def read_file(fname):
            return pd.read_csv(fname, compression='gzip')
    elif database_extension == 'csv.gz' and use_dask:
        def read_file(fname):
            return dd.read_csv(fname, compression='gzip', blocksize=None)

    data = []
    #print(FileNumbers)
    for ifile in tqdm(FileNumbers, desc=desc, leave=True, disable=not show_progress):
        
        fname = os.path.join(path2files, dataname+"{}.".format(ifile)+database_extension)
        if os.path.exists(fname):
            subdf = read_file(fname)

        if callable(prefunc2apply):
            subdf = prefunc2apply(subdf)

        if isinstance(columns, list):
            subdf = subdf[columns]

        if isinstance(dropna, list):
            subdf = subdf.dropna(subset = dropna, how = 'any')

        if isinstance(filter_dict, dict):
            if len(filter_dict) > 0 and use_dask:
                raise NotImplementedError('Dask does not currently support filtering when loading dataframes.')

            for isinkey, isinlist in filter_dict.items():
                subdf = subdf.loc[isin_sorted(subdf[isinkey].values, isinlist)]

        if isinstance(duplicate_subset, list):
            subdf = subdf.drop_duplicates(subset = duplicate_subset, keep = duplicate_keep)

        if keep_source_file:
            subdf['filetag'] = ifile

        if callable(postfunc2apply):
            subdf = postfunc2apply(subdf)

        data.append(subdf)

    if use_dask:
        data = dd.multi.concat(data)
    else:
        data = pd.concat(data)

    if isinstance(duplicate_subset, list):
        data = data.drop_duplicates(subset = duplicate_subset, keep = duplicate_keep)

    data.name = dataname
    
    return data

def append_to_preprocessed(newdf, path2database, preprocessname, startcount=0):
    """
        Append the newdf to the preprocessed DataFrames from a preprocessed directory.

        Parameters
        ----------
        :param newdf : DataFrame
            The new DataFrame to save on the processed data.  It must have at least one column in common.

        :param path2database : str
            The path to the database directory.

        :param preprocessname : str
            The type of preprocessed data to which we append the new data.

    """

    path2files = os.path.join(path2database, preprocessname)

    Nfiles = sum(preprocessname in fname for fname in os.listdir(path2files))

    for ifile in range(Nfiles):
        datadf = pd.read_hdf(os.path.join(path2files, preprocessname + '{}.hdf'.format(ifile+startcount)))
        datadf = datadf.merge(newdf, how = 'left')
        datadf.to_hdf(os.path.join(path2files, preprocessname + '{}.hdf'.format(ifile+startcount)), key = preprocessname, mode = 'w')



import os
import sys
import json
import gzip

import pandas as pd
import numpy as np
from nameparser import HumanName
import requests
from lxml import etree
from io import BytesIO

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


from pyscisci.datasource.readwrite import load_preprocessed_data, load_int, load_float, load_html_str
from pyscisci.database import BibDataBase

class DBLP(BibDataBase):
    """
    Base class for DBLP interface.

    The DBLP comes as a single xml file.  It can be downloaded from [DBLP](https://dblp.uni-trier.de/) via `donwload_from_source`

    There is no citation information!

    """


    def __init__(self, path2database = '', database_extension='csv.gz', keep_in_memory = False, global_filter=None, 
        enable_dask=False, show_progress=True):

        self._default_init(path2database, database_extension, keep_in_memory, global_filter, enable_dask, show_progress)

        self.PublicationIdType = int
        self.AffiliationIdType = int
        self.AuthorIdType = str
        self.JournalIdType=str

        self.path2journal = self.path2pub
        self.path2paa = 'publicationauthor'

    def _blank_dblp_publication(self, PublicationId = 0):
        record = {}
        record['PublicationId'] = PublicationId
        record['Title'] = ''
        record['Year'] = 0
        record['Volume'] = 0
        record['Number'] = ''
        record['Pages'] = ''
        record['JournalId'] = ''
        record['EE'] = ''
        record['TeamSize'] = 0
        record['Month'] = 1
        record['DocType'] = ''
        return record

    def _save_dataframes(self, ifile, publication, author, author_columns, author2pub):

        publication = pd.DataFrame(publication)
        publication['PublicationId'] = publication['PublicationId'].astype(int)
        publication['Year'] = publication['Year'].astype(int)
        publication['Volume'] = pd.to_numeric(publication['Volume'])
        publication['TeamSize'] = publication['TeamSize'].astype(int)
        fname = os.path.join(self.path2database, self.path2pub, '{}{}.{}'.format(self.path2pub, ifile, self.database_extension))
        self.save_data_file(publication, fname, key =self.path2pub)


        author = pd.DataFrame(author, columns = author_columns)
        author['AuthorId'] = author['AuthorId'].astype(int)
        fname = os.path.join(self.path2database, self.path2author, '{}{}.{}'.format(self.path2author, ifile, self.database_extension))
        self.save_data_file(author, fname, key =self.path2author)

        author2pub = pd.DataFrame(author2pub, columns = ['PublicationId', 'AuthorId', 'AuthorSequence'], dtype=int)
        fname = os.path.join(self.path2database, self.path2paa, '{}{}.{}'.format(self.path2paa, ifile, self.database_extension))
        self.save_data_file(author2pub, fname, key =self.path2paa)

    def preprocess(self, xml_file_name = 'dblp.xml.gz', process_name=True, num_file_lines=10**6, show_progress=True):
        """
        Bulk preprocess of the DBLP raw data.

        Parameters
        ----------
        process_name: bool, default True
            If True, then when processing the raw file, the package `NameParser <https://nameparser.readthedocs.io/en/latest/>`_
            will be used to split author FullNames.

        xml_file_name: str, default 'dblp.xml.gz'
            The xml file name.

        num_file_lines: int, default 10**6
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.

        show_progress: bool, default True
            Show progress with processing of the data.

        """

        ACCEPT_DOCTYPES = set(['article', 'inproceedings', 'proceedings', 'book', 'incollection', 'phdthesis', 'mastersthesis'])
        REJECT_DOCTYPES = set(['www'])
        DATA_ITEMS = ['title', 'booktitle', 'year', 'journal', 'ee',' url', 'month', 'mdate', 'isbn', 'publisher']
        SKIP_FIELDS = ['note', 'cite', 'cdrom', 'crossref', 'editor',  'series', 'tt', 'school', 'chapter', 'address']

        doctype = {'Article': 'j', 'Book Review':'br', 'Letter':"l", 'Review':"rev", 'Correction':"corr",
       'Editorial Material':"editorial", 'News Item':"news", 'Bibliography':"bib",
       'Biographical-Item':"bibi", 'Hardware Review':"hr", 'Meeting Abstract':"ma", 'Note':"note",
       'Discussion':"disc", 'Item About an Individual':"indv", 'Correction, Addition':"corradd",
       'Chronology':"chrono", 'Software Review':"sr", 'Reprint':"re", 'Database Review':"dr", 
       "Journal":'j', 'Book':'b', '':'', 'BookChapter':'bc', 'Conference':'c', 'Dataset':'d', 'Patent':'p', 'Repository':'r', 'Thesis':'t',
       'article': 'j', 'book':'b', '':'', 'phdthesis':'t', 'proceedings':'c', 'inproceedings':'c',
        'mastersthesis':'mst', 'incollection':'coll'}

        html_format_keys = ['<sub>', '</sub>', '<sup>', '</sup>', '<i>', '</i>']

        if show_progress:
            print("Starting to preprocess the DBLP database.")

        for hier_dir_type in [self.path2pub, self.path2author, self.path2paa]:

            if not os.path.exists(os.path.join(self.path2database, hier_dir_type)):
                os.mkdir(os.path.join(self.path2database, hier_dir_type))

        publication = []
        author = []
        author2pub = []
        journal = []

        PublicationId = 1
        AuthorId = 1
        aname2aid = {}
        author_columns = ['AuthorId', 'FullName']
        if process_name:
            author_columns += ['LastName', 'FirstName', 'MiddleName']
        JournalId = 1
        jname2jid = {}


        pub_record = self._blank_dblp_publication(PublicationId)
        pub_authors = []
        AuthorCount = 0

        ifile = 0

        # read dtd - this takes
        path2database = self.path2database # remove self to use inside of this class
        class DTDResolver(etree.Resolver):
            def resolve(self, system_url, public_id, context):
                return self.resolve_filename(os.path.join(path2database, system_url), context)

        if '.gz' in xml_file_name:
            with gzip.open(os.path.join(self.path2database, xml_file_name), 'r') as infile:
                xml_file = infile.read()

        else:
            with open(os.path.join(self.path2database, xml_file_name), 'r') as infile:
                xml_file = infile.read().encode('latin1')

        # extract the desired fields from the XML tree  #
        bytesxml = BytesIO(xml_file)
        xmltree = etree.iterparse(bytesxml, load_dtd=True, resolve_entities=True)
        xmltree.resolvers.add(DTDResolver())

        if show_progress:
            print("Xml tree parsed, iterating through elements.")

        last_position = 0
        xml_size = bytesxml.getbuffer().nbytes
        with tqdm(total=xml_size, unit='iB', unit_scale=True, desc='dblp.xml', leave=True, disable=not show_progress) as pbar:
            for event, elem in xmltree:
                if elem.tag == 'title' or elem.tag == 'booktitle':
                    pub_record['Title'] = load_html_str(elem.text)

                elif elem.tag == 'year':
                    pub_record['Year'] = load_int(elem.text)

                elif elem.tag == 'month':
                    pub_record['Month'] = load_int(elem.text)

                elif elem.tag == 'volume':
                    pub_record['Volume'] = load_int(elem.text)

                elif elem.tag == 'number':
                    pub_record['Number'] = load_html_str(elem.text)

                elif elem.tag == 'pages':
                    pub_record['Pages'] = load_html_str(elem.text)

                elif elem.tag == 'journal':
                    pub_record['JournalId'] = load_html_str(elem.text)

                elif elem.tag == 'url':
                    pub_record['URL'] = load_html_str(elem.text)

                elif elem.tag == 'ee':
                    pub_record['EE'] = load_html_str(elem.text)

                elif elem.tag == 'author':
                    AuthorCount += 1
                    fullname = load_html_str(elem.text)
                    if aname2aid.get(fullname, None) is None:
                        if process_name:
                            fullname = ''.join([i for i in fullname if not i.isdigit()]).strip()
                            hname = HumanName(fullname)
                            author.append([AuthorId, fullname, hname.last, hname.first, hname.middle])
                        else:
                            author.append([AuthorId, fullname])
                        aname2aid[fullname] = AuthorId
                        AuthorId += 1

                    pub_authors.append([PublicationId, aname2aid[fullname], AuthorCount])

                elif elem.tag in ACCEPT_DOCTYPES:
                    pub_record['TeamSize'] = AuthorCount
                    pub_record['DocType'] = doctype[load_html_str(elem.tag)]

                    publication.append(pub_record)
                    author2pub.extend(pub_authors)
                    PublicationId += 1
                    pub_record = self._blank_dblp_publication(PublicationId)
                    AuthorCount = 0
                    pub_authors = []

                    # update progress bar
                    pbar.update(bytesxml.tell() - last_position)
                    last_position = bytesxml.tell()

                    if num_file_lines > 0 and (PublicationId % num_file_lines) == 0:

                        self._save_dataframes(ifile, publication, author, author_columns, author2pub)

                        ifile += 1

                        publication = []
                        author = []
                        author2pub = []

                elif elem.tag in REJECT_DOCTYPES:
                    # the record was from a rejected category so reset record
                    pub_record = self._blank_dblp_publication(PublicationId)
                    AuthorCount = 0
                    pub_authors = []

                elif elem.tag in SKIP_FIELDS:
                    pass

        del xmltree

        self._save_dataframes(ifile, publication, author, author_columns, author2pub)




    def download_from_source(self, source_url='https://dblp.uni-trier.de/xml/', xml_file_name = 'dblp.xml.gz',
        dtd_file_name = 'dblp.dtd', show_progress=True):
        """
        Download the DBLP raw xml file and the dtd formating information from [DBLP](https://dblp.uni-trier.de/).
            1. dblp.xml.gz - the compressed xml file
            2. dblp.dtd - the dtd containing xml syntax

        The files will be saved to the path specified by `path2database`.

        Parameters
        ----------
        source_url: str, default 'https://dblp.uni-trier.de/xml/'
            The base url from which to download.

        xml_file_name: str, default 'dblp.xml.gz'
            The xml file name.

        dtd_file_name: str, default 'dblp.dtd'
            The dtd file name.

        show_progress: bool, default True
            Show progress with processing of the data.

        """

        block_size = 1024 #1 Kibibyte

        req_stream = requests.get(os.path.join(source_url, xml_file_name), stream=True)
        total_size = int(req_stream.headers.get('content-length', 0))

        if not os.path.exists(self.path2database):
            os.mkdir(self.path2database)

        with tqdm(total=total_size, unit='iB', unit_scale=True, desc='dblp.xml.gz', leave=True, disable=not show_progress) as pbar:
            with open(os.path.join(self.path2database, xml_file_name), "wb") as outfile:

                for block in req_stream.iter_content(block_size):
                    outfile.write(block)

                    # update progress bar
                    pbar.update(len(block))

        with open(os.path.join(self.path2database, dtd_file_name), 'w') as outfile:
            outfile.write(requests.get(os.path.join(source_url, dtd_file_name)).content.decode('latin1'))

    def parse_affiliations(self, preprocess = False):
        raise NotImplementedError("DBLP is stored as a single xml file.  Run preprocess to parse the file.")

    def parse_authors(self, preprocess = False, process_name = True, num_file_lines = 5*10**6):
        raise NotImplementedError("DBLP is stored as a single xml file.  Run preprocess to parse the file.")

    def parse_publications(self, preprocess = False, num_file_lines=10**7):
        raise NotImplementedError("DBLP is stored as a single xml file.  Run preprocess to parse the file.")

    def parse_references(self, preprocess = False, num_file_lines=10**7):
        raise NotImplementedError("DBLP does not contain reference or citation information.")

    def parse_publicationauthoraffiliation(self, preprocess = False, num_file_lines=10**7):
        raise NotImplementedError("DBLP is stored as a single xml file.  Run preprocess to parse the file.")

    def parse_fields(self, preprocess = False, num_file_lines=10**7):
        raise NotImplementedError("DBLP does not contain field information.")

    @property
    def author2pub(self):
        """
        The DataFrame keeping all publication, author relationships.  Columns may depend on the specific datasource.

        Notes
        -------
        columns: 'PublicationId', 'AuthorId', 'AuthorOrder'

        """
        if self._author2pub is None:
            if self.keep_in_memory:
                self._author2pub = self.load_publicationauthor(show_progress=self.show_progress)
            else:
                return self.load_publicationauthor(show_progress=self.show_progress)

        return self._author2pub

    def load_publicationauthor(self, preprocess = True, columns = None, filter_dict = None, duplicate_subset = None,
        duplicate_keep = 'last', dropna = None, show_progress=False):
        """
        Load the PublicationAuthor DataFrame from a preprocessed directory.  For DBLP, you must run preprocess before
        the dataframe is available for use.

        Parameters
        ----------
        preprocess : bool, default True, Optional
            Attempt to load from the preprocessed directory.

        columns : list, default None, Optional
            Load only this subset of columns

        filter_dict : dict, default None, Optional
            Dictionary of format {"ColumnName":"ListofValues"} where "ColumnName" is a data column
            and "ListofValues" is a sorted list of valid values.  A DataFrame only containing rows that appear in
            "ListofValues" will be returned.

        duplicate_subset : list, default None, Optional
            Drop any duplicate entries as specified by this subset of columns

        duplicate_keep : str, default 'last', Optional
            If duplicates are being dropped, keep the 'first' or 'last'
            (see `pandas.DataFram.drop_duplicates <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html>`_)

        dropna : list, default None, Optional
            Drop any NaN entries as specified by this subset of columns

        Returns
        -------
        DataFrame
            PublicationAuthor DataFrame.

        """
        if show_progress:
            show_progress='Loading PublicationAuthor'
        if preprocess and os.path.exists(os.path.join(self.path2database, 'publicationauthor')):
            return load_preprocessed_data('publicationauthor', path2database=self.path2database, columns=columns,
                filter_dict=filter_dict, duplicate_subset=duplicate_subset, duplicate_keep=duplicate_keep, dropna=dropna,
                show_progress=show_progress)
        else:
            raise NotImplementedError("DBLP is stored as a single xml file.  Run preprocess to parse the file.")

    def load_journals(self, preprocess = True, columns = None, filter_dict = {}, duplicate_subset = None,
        duplicate_keep = 'last', dropna = None, prefunc2apply=None, postfunc2apply=None, show_progress=False):

        raise NotImplementedError("The DBLP does not have prespecified journal information.")




import datetime
import numpy as np

import matplotlib.pylab as plt
import matplotlib.colors as colors

from pyscisci.methods.productivitytrajectory import piecewise_linear

def career_impacttimeline(impact, datecol = 'Date', impactcol='Ctotal', fill_color='orange', hot_streak_info = None, 
    edge_color='k', streak_color='firebrick', ax=None):

    if ax is None:
        fig, ax = plt.subplots(1,1,figsize=(10,6))

    if isinstance(datecol, str):
        dates = impact[datecol].values
    elif isinstance(datecol, np.ndarray):
        dates = datecol

    if isinstance(impactcol, str):
        cites = impact[impactcol].values
    elif isinstance(impactcol, np.ndarray):
        cites = impactcol

    for d, c in zip(dates, cites):
        if isinstance(d, str):
            if 'T' in d:
                d= datetime.datetime.strptime(d.split('T')[0], '%Y-%m-%d')
            else:
                d= datetime.datetime.strptime(d.split(' ')[0], '%Y-%m-%d')

        ax.plot([d]*2, [0,c], c='k', lw = 0.5) 
        ax.scatter(d, c, color=fill_color, edgecolor=edge_color, linewidth=0.5, zorder=100)
    
    if not hot_streak_info is None:
        nstreaks = 2 - hot_streak_info[:3].isnull().sum()
        
        if nstreaks == 1:
            gamma0, gamma1 = 10**(hot_streak_info[:2])
            streak_start, streak_end = hot_streak_info[3:5].astype(int)
            ax.plot(datecol[:streak_start], [gamma0]*streak_start, c=streak_color)
            ax.plot(datecol[streak_start:(streak_end+1)], [gamma1]*(streak_end-streak_start+1), c=streak_color)
            ax.plot(datecol[(streak_end+1):], [gamma0]*(datecol.shape[0]-streak_end-1), c=streak_color)

    return ax


def career_productivitytimeline(yearlyprod, productivity_trajectory = None, datecol = 'Year', fill_color='blue', ax=None):

    if ax is None:
        fig, ax = plt.subplots(1,1,figsize=(10,6))

    ax.bar(yearlyprod[datecol].values, yearlyprod['YearlyProductivity'].values, color=fill_color) 
    
    if not productivity_trajectory is None:
        t_break, b, m1, m2 = productivity_trajectory[['t_break','b','m1','m2']].values[0]

        ts = np.arange(yearlyprod[datecol].min(), yearlyprod[datecol].max()+1)
        ax.plot(ts, piecewise_linear(ts, t_break, b, m1, m2), color='black')
        
    return ax

def hex2rgb(value):
    value = value.lstrip('#')
    lv = len(value)
    return tuple(int(value[i:i + lv // 3], 16)/255. for i in range(0, lv, lv // 3))

def hex2rgba(value, alpha = 1):
    return hex2rgb(value) + (alpha,)

class MidpointNormalize(colors.Normalize):
    """
    Normalise the colorbar so that diverging bars work there way either side from a prescribed midpoint value)

    e.g. im=ax1.imshow(array, norm=MidpointNormalize(midpoint=0.,vmin=-100, vmax=100))
    """
    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        self.midpoint = midpoint
        colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y), np.isnan(value))


# -*- coding: utf-8 -*-
"""
.. module:: embedding
    :synopsis: The main Embedding Class

.. moduleauthor:: Jisung Yoon <jisung.yoon92@gmail.com>
 """
import json
import sys

import numpy as np
from joblib import Parallel, delayed
from scipy import sparse, spatial

# For now, I dont think we need to make the full pySciSci package dependent on these packages
try:
    import numba
except ImportError:
    raise ImportError(
        "Please install numba to take full advantage of fast embedding tools. \n pip install numba"
    )

try:
    from gensim.models import Word2Vec
except ImportError:
    raise ImportError(
        "Please install gensim to take full advantage of fast embedding tools. \n pip install gensim"
    )

try:
    import networkx as nx
except ImportError:
    raise ImportError(
        "Please install networkx if you want to take an input as networkx object. \n pip install networkx"
    )


class Node2Vec(object):
    """
    Node2Vec node embedding object
    Paper: https://snap.stanford.edu/node2vec/
    This code is originated from https://github.com/yy/residual-node2vec written by Sadamori Kojaku,
    and modified for its own purpose.
    
    Node2vec works as follows.
    1. Given the network, it generates random walkers on the network.
    2. Random walkers start from each node and travel around the network until the length of the random-walker reaches walk_length. (Total num_walks * walk_length)
    3. Get the embedding of each node with passing these sentences into Gensim Word2vec package (SGNS model).
    """

    def __init__(
        self,
        A,
        entity2int,
        num_walks=10,
        walk_length=80,
        p=1.0,
        q=1.0,
        dimensions=128,
        window_size=10,
        epoch=1,
        workers=1,
    ):
        """
        :param A: Sparse Adjacecny Matrix
        :param entity2int: mapping dict for the network
        :param num_walks: Number of random walker per node
        :param walk_length: Length(number of nodes) of random walker
        :param p: the likelihood of immediately revisiting a node in the walk
        :param q: search to differentiate between “inward” and “outward” nodes in the walk
        :param dimensions: Dimension of embedding vectors
        :param window_size: Maximum distance between the current and predicted node in the network
        :param epoch: Number of epochs over the walks
        :param workers: Number of CPU cores that will be used in training
        """

        self.A = to_csr_adjacency_matrix(
            A
        )  # transform input to csr_matrix for calculation
        self.entity2int = entity2int

        # parameters for random walker
        self.num_walks = num_walks
        self.walk_length = walk_length
        self.p = p
        self.q = q

        # parameters for learning embeddings
        self.dimensions = dimensions
        self.window_size = window_size
        self.epoch = epoch

        # computing configuration and path
        self.workers = workers
        self.by_pass_mode = (
            True if p == q == 1 else False
        )  # for the random-walker with p=q=1, which is an unbiased random walker

        self.simulate_walks()

    def simulate_walks(self):
        self.num_nodes = self.A.shape[0]

        if self.by_pass_mode:
            self.walks = simulate_walk(
                self.A,
                num_walk=self.num_walks,
                walk_length=self.walk_length,
                start_node_ids=np.arange(self.num_nodes),
                workers=self.workers,
            )

        else:
            Asupra, node_pairs = construct_line_net_adj(self.A)

            # Find the starting node ids
            start_node_ids = np.where(node_pairs[:, 0] == node_pairs[:, 1])[0]
            # Simulate walks
            trajectories = simulate_walk(
                Asupra,
                num_walk=self.num_walks,
                walk_length=self.walk_length,
                start_node_ids=start_node_ids,
                workers=self.workers,
            )

            # Convert supra-node id to the id of the nodes in the original net
            self.walks = node_pairs[trajectories.reshape(-1), 1].reshape(
                trajectories.shape
            )
        self.walks = self.walks.astype(str).tolist()

    def learn_embedding(self):
        """
        Learning an embedding of nodes in the base graph.
        :return self.embedding: Embedding of nodes in the latent space.
        """
        self.model = Word2Vec(
            self.walks,
            size=self.dimensions,
            window=self.window_size,
            min_count=0,
            sg=1,
            workers=self.workers,
            iter=self.epoch,
        )
        self.embedding = {
            entity: self.model.wv[str(index)]
            for entity, index in self.entity2int.items()
        }

        return self.embedding

    def save_embedding(self, file_name):
        """
        :param file_name: name of file_name
        """
        self.model.wv.save_word2vec_format("{}.w2v".format(file_name))
        with open("{}_mapper.json", format(file_name), "w") as fp:
            json.dump(self.entity2int, fp)


# utility function for convert the type
def to_csr_adjacency_matrix(net):
    if sparse.issparse(net):
        if type(net) == "scipy.sparse.csr.csr_matrix":
            return net
        return sparse.csr_matrix(net)
    elif "networkx" in "%s" % type(net):
        return nx.adjacency_matrix(net)
    elif "numpy.ndarray" == type(net):
        return sparse.csr_matrix(net)


# fast simulate walk
def simulate_walk(
    A, num_walk, walk_length, start_node_ids, workers=1,
):
    """
    Simulate PageRank random walk on the given network.
    The function is compatible with the weighted and directed networks.
    We exploit the sparse.sparse.csr_matrix
    data structure to speed up the simulation of random walkers.
    See `_csr_walk` for details.
    Parameters
    ----------
    A : scipy.sparse.csr_matrix
        Adjacency matrix for the network.
        Self-loops must be removed.
    num_walk: int
        Number of walkers
    walk_length : int
        Number of steps for a single walker
    workers : int
        Number of CPUs
    Return
    ------
    trajectories : numpy.ndarray
        Trajectory of random walker.
        Each row indicates the sequence of nodes that a walker visits.
    """

    if A.getformat() != "csr":
        raise TypeError("A should be in the scipy.sparse.csr_matrix")

    # Extract the information on the csr matrix
    A_indptr = A.indptr
    A_indices = A.indices
    A_data = A.data.astype(np.float64)

    # Compute the transition probability priori
    num_nodes = A.shape[0]
    A_data = _calc_cumulative_trans_prob(A_indptr, A_data, num_nodes)

    if start_node_ids is None:
        start_node_ids = np.arange(num_nodes)

    # Parallelize the process
    if workers == 1:
        trajectories = [
            _csr_walk(A_indptr, A_indices, A_data, start_node_ids, walk_length,)
            for i in range(num_walk)
        ]
    else:
        trajectories = Parallel(n_jobs=workers)(
            delayed(_csr_walk)(
                A_indptr, A_indices, A_data, start_node_ids, walk_length,
            )
            for i in range(num_walk)
        )
    return np.vstack(trajectories)


@numba.jit(nopython=True, cache=True)
def _csr_walk(
    A_indptr, A_indices, A_data, start_node_ids, walk_length,
):
    """
    Simulate random walk on the given network.
    We exploit the scipy.sparse.csr_matrix data structure to speed
    up the simulation of random walks. Specifically, consider
    scipy.sparse.csr_matrix, A, representing the adjacency matrix.
    A has the following 3 attributes:
    - A.indices
    - A.data
    - A.indptr
    A.indices and A.data store the column IDs and the value of
    the non-zero element, respectively. A.indptr[i] is the first
    index of the A.indices for the ith row. One can find the neighbors
    of node i by A.indices[A.indptr[i]:A.indptr[i+1]] and the edge weight
    by A.data[A.indptr[i]:A.indptr[i+1]]
    The data structure is memory efficient and we exploit the efficiency
    to speed up the random walk process.
    Furthermore, we use `numba` to further speed up the random walk process.
    Parameters
    ----------
    A_indptr : numpy.ndarray
        A_indptr is given by A.indptr, where A is the scipy.sparse.csr_matrix
        representing the adjacency matrix for the original network
    A_indices : numpy.ndarray
        A_indicesr is given by A.indices, where A is the
        scipy.sparse.csr_matrix as in A_indptr
    A_data : numpy.ndarray
        A_data is given by A.data, where A is the
        scipy.sparse.csr_matrix as in A_indptr
    start_node_ids : numpy.ndarray
        ID of the nodes where random walker starts.
    walk_length : int
        Length of a single walk
    restart_prob : float
        Restart probability
    ergodic_proces : bool 
        If set True, the random walker will already teleport back to the starting node when it hits 
        a dangling node (i.e., node with no outgoing edge). If set False, the random walk ends at dangling nodes.
    Return
    ------
    L : numpy.ndarray
        Size (satrt_node_ids.size, walk_length).
        L[i,t] indicates the node at which the walker 
        starting from start_node_ids[i] visits at time t
    """

    if start_node_ids.size == 0:
        return np.zeros((1, 1), dtype=np.int32)
    # Simulate the random walk
    L = -np.ones((start_node_ids.size, walk_length), dtype=np.int32)
    for sample_id, start_node in enumerate(start_node_ids):

        visit = start_node

        L[sample_id, 0] = visit
        for t in range(1, walk_length):

            outdeg = A_indptr[visit + 1] - A_indptr[visit]

            # If reaches to an absorbing state, finish the walk
            if outdeg == 0:
                break
            else:
                # find a neighbor
                _next_node = np.searchsorted(
                    A_data[A_indptr[visit] : A_indptr[visit + 1]],
                    np.random.rand(),
                    side="right",
                )
                next_node = A_indices[A_indptr[visit] + _next_node]

            # Record the transition
            L[sample_id, t] = next_node

            # Move
            visit = next_node
    return L


@numba.jit(nopython=True)
def _calc_cumulative_trans_prob(A_indptr, A_data, num_nodes):
    """
    Calculate the transition probability for the random walks on
    the given network. The probability is expressed as the
    cumulative probability mass for a fast random
    drawing of a neighboring node.
    Parameters
    ----------
    A_indptr : numpy.ndarray
        A_indptr is given by A.indptr, where A is the scipy.sparse.csr_matrix
        representing the adjacency matrix for the original network
    A_data : numpy.ndarray
        A_data is given by A.data, where A is the scipy.sparse.csr_matrix
        as in A_indptr
    num_nodes : int
        number of nodes in the network
    Return
    ------
    L : numpy.ndarray
        Size (satrt_node_ids.size, walk_length).
        L[i,t] indicates the node at which the walker
        starting from start_node_ids[i] visits at time t
    """
    for i in numba.prange(num_nodes):

        # Compute the out-deg
        outdeg = np.sum(A_data[A_indptr[i] : A_indptr[i + 1]])

        # Compute the transition probability of walkers
        # we compute the cumulative of the transition probabilities, i.e.,
        # A_data[i,j] is the (cumulative) probability of the edge.
        # sum_{j}^ N A[i,j] = 1.
        A_data[A_indptr[i] : A_indptr[i + 1]] = np.cumsum(
            A_data[A_indptr[i] : A_indptr[i + 1]]
        ) / np.maximum(outdeg, 1e-8)

    return A_data


# This function uses the following subroutines:
# - _construct_node2vec_supra_net_edge_pairs
def construct_line_net_adj(A, p=1, q=1, add_source_node=True):
    """
    Construct the supra-adjacent matrix for the weighted network.
    The random walk process in the Node2Vec is the second order Markov process,
    where a random walker at node i remembers the previously visited node j
    and determines the next node k based on i and j.
    We transform the 2nd order Markov process to the 1st order Markov
    process on supra-nodes. Each supra-node represents a pair of
    nodes (j,i) in the original network, a pair of the previouly
    visited node and current node. We place an edge
    from a supra-node (j,i) to another supra-node (i,k) if the random walker
    can transit from node j, i and to k. The weight of edge is given by the
    unnormalized transition probability for the node2vec random walks.
    """

    if A.getformat() != "csr":
        raise TypeError("A should be in the scipy.sparse.csc_matrix")

    # Preprocessing
    # We efficiently simulate the random walk process using the scipy.csr_matrix
    # data structure. The followings are the preprocess for the simulation.
    A.sort_indices()  # the indices have to be sorted
    A_indptr_csr = A.indptr.astype(np.int32)
    A_indices_csr = A.indices.astype(np.int32)
    A_data_csr = A.data.astype(np.float32)
    A = A.T
    A.sort_indices()  # the indices have to be sorted
    A_indptr_csc = A.indptr.astype(np.int32)
    A_indices_csc = A.indices.astype(np.int32)
    A_data_csc = A.data.astype(np.float32)
    num_nodes = A.shape[0]

    # Make the edge list for the supra-networks
    supra_edge_list, edge_weight = construct_node2vec_supra_net_edge_pairs(
        A_indptr_csr,
        A_indices_csr,
        A_data_csr,
        A_indptr_csc,
        A_indices_csc,
        A_data_csc,
        num_nodes,
        p=p,
        q=q,
        add_source_node=add_source_node,
    )

    # Remove isolated nodes from the supra-network and
    # re-index the ids.
    supra_nodes, supra_edge_list = np.unique(supra_edge_list, return_inverse=True)
    supra_edge_list = supra_edge_list.reshape((edge_weight.size, 2))

    # Each row indicates the node pairs in the original net that constitute the supra node
    src_trg_pairs = (np.vstack(divmod(supra_nodes, num_nodes)).T).astype(int)

    # Construct the supra-adjacency matrix
    supra_node_num = supra_nodes.size
    Aspra = sparse.csr_matrix(
        (edge_weight, (supra_edge_list[:, 0], supra_edge_list[:, 1])),
        shape=(supra_node_num, supra_node_num),
    )
    return Aspra, src_trg_pairs


def sem_axis(emb, positive_entities, negative_entities):
    """
    SemAxis Code
    Paper: https://arxiv.org/abs/1806.05521
    SemAxis is a technique that leverages the latent semantic characteristics of word embeddings to represent 
    The position of terms along a conceptual axis, reﬂecting the relationship of these terms to the concept.
    
    Semaxis works as follows.
    1. Given the set of positive instances, S+, and set of negative instances S-, calculate the average vector of each set (V+, V-).
    2. calculate the semantic axis can be calculated as V_{axis} = V+ - V-, and project another instance into this semantic axis with measuring cosine similarity to the semantic axis.
    
    Higher scores mean that organization w is more closely aligned to S+ than S− .
    """
    mean_positivie_vec = np.array([emb[entity] for entity in positive_entities])
    mean_negative_vec = np.array([emb[entity] for entity in negative_entities])

    target_axis = mean_positivie_vec - mean_negative_vec
    sem_axis_result_dict = {
        k: 1 - spatial.distance.cosine(target_axis, v) for k, v in emb.items()
    }

    return sem_axis_result_dict



# -*- coding: utf-8 -*-
"""
.. module:: sleepingbeauty
    :synopsis: Calculate the sleeping beauty coefficient.

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import sys

import pandas as pd
import numpy as np

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.utils import zip2dict


## Beauty-coefficient
def beauty_coefficient(c):
    """
    This function calculates the sleeping beauty coefficient and awakening time for a publication.  See :cite:`ke2015beauty` for details.

    Parameters
    ----------
    c : numpy array
        The yearly citation counts for the publication.

    Returns
    ----------
    B : float
        Sleeping Beauty Coefficient

    t_a : int
        The awakening time

    """
    c = c.values
    t_m = np.argmax(c)
    B_denom = c
    B_denom[c==0] = 1

    # :cite:`ke2015beauty` eq 1/2
    l_t = ((c[t_m] - c[0])/t_m *np.arange(c.shape[0]) + c[0] - c)/B_denom

    # :cite:`ke2015beauty` eq 2
    B = l_t[:(t_m+1)].sum()

    d_denom = np.sqrt((c[t_m] - c[0])**2 + t_m**2)
    d_t = np.abs( (c[t_m] - c[0]) * np.arange(c.shape[0]) + t_m * (c[0] - c)) / d_denom

    # :cite:`ke2015beauty` eq 3
    t_a = np.argmax(d_t[:(t_m+1)])

    return pd.Series([B, t_a], index=['BeautyCoefficient', 'Awakening'])

def compute_sleepingbeauty(df, colgroupby, colcountby, coldate='Year', show_progress=False):
    """
    Calculate the sleeping beauty and awakening time for each group in the DataFrame.  See :cite:`ke2015beauty` for details.

    The algorithmic implementation for each publication can be found in :py:func:`sleepingbeauty.beauty_coefficient`.

    Parameters
    ----------
    df : DataFrame
        A DataFrame with the citation information for each publication in each year.

    colgroupby : str
        The DataFrame column with Publication Ids.

    colcountby : str
        The DataFrame column with Citation counts for each publication.

    coldate : str
        The DataFrame column with Year information.

    Returns
    -------
    DataFrame
        DataFrame with 3 columns: colgroupby, 'Beauty' and 'Awakening'

        """
    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc='Beauty', disable= not show_progress)

    def fill_missing_dates(subdf):
        subdf = subdf.set_index(coldate).reindex(np.arange(subdf[coldate].min(), subdf[coldate].max()+1)).fillna(0).reset_index()
        return subdf

    # first fill in missing dates
    df = df.groupby(colgroupby, sort=False, group_keys=False).apply(fill_missing_dates)

    #get start year
    syear = df.groupby(colgroupby, sort=False)[coldate].min()

    # now find the beauty coefficient and awakening year
    beauty = df.groupby(colgroupby, sort=False)[colcountby].progress_apply(beauty_coefficient).unstack(1).reset_index()

    # translate the awakening from index to year
    beauty['Awakening'] = [a+syear[pid] for pid,a in beauty[[colgroupby, 'Awakening']].values]
    
    return beauty


import os
import sys
import json
import gzip
import zipfile

import pandas as pd
import numpy as np
from nameparser import HumanName

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.datasource.readwrite import load_preprocessed_data, load_int, load_float, load_html_str
from pyscisci.database import BibDataBase
from pyscisci.utils import download_file_from_google_drive

# hide this annoying performance warnings
import warnings
warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)


class APS(BibDataBase):
    """
    Base class for APS interface.

    The APS comes as a single xml file.

    You must request usage through their website: https://journals.aps.org/datasets

    """

    def __init__(self, path2database = '', database_extension='csv.gz', keep_in_memory = False, global_filter=None, 
        enable_dask=False, show_progress=True):

        self._default_init(path2database, database_extension, keep_in_memory, global_filter, enable_dask, show_progress)

        self.PublicationIdType = str
        self.AffiliationIdType = str
        self.JournalIdType = str

        


    def preprocess(self, archive_year=2019, pubid2int=False, metadata_archive=None, citation_archive=None, show_progress=True):
        """
        Bulk preprocess the APS raw data.

        """
        if metadata_archive is None:
            metadata_archive = 'aps-dataset-metadata-{}.zip'.format(archive_year)

        self.parse_publications(preprocess=True, preprocess_dicts=True, pubid2int=pubid2int,
            archive_name = metadata_archive, show_progress=show_progress)

        if citation_archive is None:
            citation_archive = 'aps-dataset-citations-{}.zip'.format(archive_year)

        self.parse_references(preprocess=True, pubid2int=pubid2int, archive_name=citation_archive, show_progress=show_progress)



    def download_from_source(self, files_to_download='all'):

        if files_to_download in ['all', 'orig']:
            import webbrowser
            webbrowser.open("https://journals.aps.org/datasets")

            print("APS is shared by request from the American Physical Society.  Contact APS to download the source files.")


        if files_to_download in ['all', 'paa_supplement']:

            aps_author_file_id = '1U6f9AYQJUHQ_IzeiI-rADFsrg7LWH3WO'

            if not os.path.exists(os.path.join(self.path2database, 'publicationauthoraffiliation2010supplement')):
                os.mkdir(os.path.join(self.path2database, 'publicationauthoraffiliation2010supplement'))

            filename = os.path.join(self.path2database, 'publicationauthoraffiliation2010supplement', 'publicationauthoraffiliation0.hdf')
            if os.path.exists(filename): os.remove(filename)
            filename = os.path.join(self.path2database, 'publicationauthoraffiliation2010supplement', 'publicationauthoraffiliation2010supplement0.hdf')

            apsauthors_gzip = download_file_from_google_drive(file_id=aps_author_file_id, destination= filename)

            self.set_new_data_path(dataframe_name='paa', new_path='publicationauthoraffiliation2010supplement')

            print("New data saved to {}.".format('publicationauthoraffiliation2010supplement'))

        if not files_to_download in ['all', 'orig', 'paa_supplement']:
            print("Unrecognized file name.")


    def parse_affiliations(self, preprocess = False, show_progress=False):
        raise NotImplementedError("APS is stored as a json archive.  Run preprocess to parse the archive.")

    def parse_authors(self, preprocess = False, process_name = True, num_file_lines = 5*10**6, show_progress=False):
        raise NotImplementedError("APS does not contain disambiguated author information.")

    def parse_publications(self, preprocess=False, preprocess_dicts=True, pubid2int=False,
        archive_name = 'aps-dataset-metadata-2019.zip', show_progress=False):

        archive = zipfile.ZipFile(os.path.join(self.path2database, archive_name), 'r')
        metadata_files = [fname for fname in archive.namelist() if 'aps-dataset-metadata' in fname and '.json' in fname]

        # check that the archive concatins the expected directory
        if len(metadata_files) > 0:

            if preprocess:

                for hier_dir_type in [self.path2pub, self.path2journal, self.path2affiliation, self.path2paa, self.path2pub2field, self.path2fieldinfo]:

                    if not os.path.exists(os.path.join(self.path2database, hier_dir_type)):
                        os.mkdir(os.path.join(self.path2database, hier_dir_type))


            journal_dict = {}
            journal_column_names = ['JournalId', 'FullName', 'AbbreviatedName', 'Publisher']

            pub_column_names = ['PublicationId', 'Title', 'Date', 'Year', 'Doi', 'JournalId', 'Volume', 'Issue', 'PageStart', 'PageEnd', 'DocType', 'TeamSize']

            pub = []
            pub2year = {}
            pub2doctype = {}
            pub2int = {}
            ipub = 0
            if pubid2int:
                pubintcol = ['PublicationId']
                self.PublicationIdType=int
            else:
                pubintcol = []

            iaff = 0
            affil_dict = {}
            paa = []

            field_dict = {}
            pub2field = []

            for fname in tqdm(metadata_files, desc='aps-metadata', leave=True, disable=not show_progress):
                # load pub json
                pubjson = json.loads(archive.read(fname).decode('utf-8'))
                ipub += 1

                # start parsing publication information
                if pubid2int:
                    pubid = ipub
                    pub2int[pubjson.get('id', '')] = pubid
                else:
                    pubid = pubjson.get('id', '')
                pubinfo = [pubid]
                pubinfo.append(pubjson.get('title', {}).get('value', ''))
                pubinfo.append(pubjson.get('date', ''))
                pubinfo.append(load_int(pubjson.get('date', '').split('-')[0]))
                pub2year[pubid] = pubinfo[-1]
                pubinfo.append(pubjson.get('id', ''))

                # journal of publication
                journalid = pubjson.get('journal', {}).get('id', '')
                pubinfo.append(journalid)
                pubinfo.append(load_int(pubjson.get('volume', {}).get('number', '')))
                pubinfo.append(load_int(pubjson.get('issue', {}).get('number', '')))

                # add pagenumber info
                pubinfo.append(load_int(pubjson.get('pageStart', '')))
                if not pubjson.get('pageEnd', None) is None:
                    pubinfo.append(load_int(pubjson.get('pageEnd', '')))

                elif not (pubjson.get('numPages', None) is None and not pubjson.get('pageStart', None) is None):
                    #pubinfo.append(pubinfo[-1] + load_int(pubjson.get('numPages', '')))
                    pubinfo.append(None)
                    
                else:
                    pubinfo.append(None)

                # add the doctype
                pubinfo.append(pubjson.get('articleType', ''))
                pub2doctype[pubid] = pubinfo[-1]

                # calculate TeamSize
                pubinfo.append(len(pubjson.get('authors', [])))

                # finish publication infor
                pub.append(pubinfo)

                # check if we need to save journal information
                if journal_dict.get(journalid, None) is None:
                    journal_dict[journalid] = pubjson.get('journal', {})
                    journal_dict[journalid]['Publisher'] = pubjson.get('rights', {}).get('copyrightHolders', [{'name':''}])[0].get('name', '')

                # start parsing affiliation information
                pub_affid_map = {}
                for pubaffdict in pubjson.get('affiliations', []):
                    # check if the affiliation has been used before (only using string match)
                    # ToDo: add disambigation
                    if affil_dict.get(pubaffdict.get('name', ''), None) is None:
                        affil_dict[pubaffdict.get('name', '')] = iaff
                        iaff += 1

                    # map the affiliation to the AffiliationId
                    pub_affid_map[pubaffdict.get('id', '')] = affil_dict[pubaffdict.get('name', '')]

                authorseq = 1
                # now start parsing author information
                for authordict in pubjson.get('authors', []):
                    for affid in authordict.get('affiliationIds', [None]):
                        paa.append([pubid, authordict.get('name', ''), pub_affid_map.get(affid, None), authorseq])

                    authorseq += 1

                classificationschemes = pubjson.get('classificationSchemes', {})

                # now do the subject classifications
                for subjectdict in classificationschemes.get('subjectAreas', []):
                    fid = subjectdict.get('id', None)

                    if not fid is None and len(fid) > 0:
                        pub2field.append([pubid, fid])

                        if field_dict.get(fid, None) is None:
                            field_dict[fid] = [subjectdict.get('label', None), 'subjectAreas']

                # now do the subject disciplines
                for subjectdict in classificationschemes.get('physh', {}).get('disciplines', []):
                    fid = subjectdict.get('id', None)

                    if not fid is None and len(fid) > 0:
                        pub2field.append([pubid, fid])

                        if field_dict.get(fid, None) is None:
                            field_dict[fid] = [subjectdict.get('label', None), 'disciplines']

                # now do the subject concepts
                for subjectdict in classificationschemes.get('physh', {}).get('concepts', []):
                    fid = subjectdict.get('id', None)

                    if not fid is None and len(fid) > 0:
                        pub2field.append([pubid, fid])

                        if field_dict.get(fid, None) is None:
                            field_dict[fid] = [subjectdict.get('label', None), 'concepts']

                    

                # ToDo: parse concepts

            if show_progress:
                print("Parsing Complete\nSaving Publication DataFrames")

            pub = pd.DataFrame(pub, columns = pub_column_names)
            for intcol in pubintcol + ['Year']:
                pub[intcol] = pub[intcol].astype(int)

            journal_rename_dict = {'name':'FullName', 'id':'JournalId', 'abbreviatedName':'AbbreviatedName'}
            journal = pd.DataFrame(journal_dict.values()).rename(columns=journal_rename_dict)

            affiliation = pd.DataFrame([[affid, name] for name, affid in affil_dict.items()], columns = ['AffiliationId', 'Address'])

            paa = pd.DataFrame(paa, columns = ['PublicationId', 'OrigAuthorName', 'AffiliationId', 'AuthorSequence'])
            for intcol in pubintcol+['AuthorSequence']:
                paa[intcol] = paa[intcol].astype(int)

            pub2field = pd.DataFrame(pub2field, columns = ['PublicationId', 'FieldId'])
            for intcol in pubintcol:
                pub2field[intcol] = pub2field[intcol].astype(int)

            field = pd.DataFrame([[fieldid] + fieldname for fieldid, fieldname in field_dict.items()], columns = ['FieldId', 'FullName', 'ClassificationType'])

            if preprocess:
                fname = os.path.join(self.path2database, self.path2pub, '{}{}.{}'.format(self.path2pub, 0, self.database_extension))
                self.save_data_file(pub, fname, key =self.path2pub)

                if pubid2int:
                    with gzip.open(os.path.join(self.path2database, 'pub2int.json.gz'), 'w') as outfile:
                        outfile.write(json.dumps(pub2int).encode('utf8'))

                if preprocess_dicts:
                    with gzip.open(os.path.join(self.path2database, 'pub2year.json.gz'), 'w') as outfile:
                        outfile.write(json.dumps(pub2year).encode('utf8'))

                    with gzip.open(os.path.join(self.path2database, 'pub2doctype.json.gz'), 'w') as outfile:
                        outfile.write(json.dumps(pub2doctype).encode('utf8'))

                fname = os.path.join(self.path2database, self.path2journal, '{}{}.{}'.format(self.path2journal, 0, self.database_extension))
                self.save_data_file(journal, fname, key =self.path2journal)

                fname = os.path.join(self.path2database, self.path2affiliation, '{}{}.{}'.format(self.path2affiliation, 0, self.database_extension))
                self.save_data_file(affiliation, fname, key =self.path2affiliation)

                fname = os.path.join(self.path2database, self.path2paa, '{}{}.{}'.format(self.path2paa, 0, self.database_extension))
                self.save_data_file(paa, fname, key =self.path2paa)

                fname = os.path.join(self.path2database, self.path2pub2field, '{}{}.{}'.format(self.path2pub2field, 0, self.database_extension))
                self.save_data_file(pub2field, fname, key =self.path2pub2field)

                fname = os.path.join(self.path2database, self.path2fieldinfo, '{}{}.{}'.format(self.path2fieldinfo, 0, self.database_extension))
                self.save_data_file(field, fname, key =self.path2fieldinfo)

        else:
            raise FileNotFoundError('The archive {0} does not contain a metadata directory: {1}.'.format(archive_name, 'aps-dataset-metadata'))

    def parse_references(self, preprocess=False, pubid2int=False, archive_name='aps-dataset-citations-2019.zip', show_progress=False):

        if preprocess and not os.path.exists(os.path.join(self.path2database, self.path2pub2ref)):
            os.mkdir(os.path.join(self.path2database, self.path2pub2ref))

        if pubid2int:
            with gzip.open(os.path.join(self.path2database, 'pub2int.json.gz'), 'r') as infile:
                pub2int = json.loads(infile.read().decode('utf8'))

            def pub2int_map(doi):
                if pub2int.get(doi, None) is None:
                    pub2int[doi] = len(pub2int) + 1

                return pub2int[doi]


        rename_dict = {'citing_doi':'CitingPublicationId', 'cited_doi':'CitedPublicationId'}

        archive = zipfile.ZipFile(os.path.join(self.path2database, archive_name), 'r')

        citation_file = [fname for fname in archive.namelist() if 'aps-dataset-citations' in fname]
        if len(citation_file) > 0:
            citation_file = citation_file[0]
            csvlines = archive.read(citation_file).decode('utf-8').split('\n')

            pub2ref = [line.split(',') for line in tqdm(csvlines, desc='aps-citations', leave=True, disable=not show_progress)]

            pub2ref = pd.DataFrame(pub2ref[1:], columns = pub2ref[0]).rename(columns=rename_dict)
            if pubid2int:
                pub2ref['CitingPublicationId'] = [pub2int_map(pid) for pid in pub2ref['CitingPublicationId'].values]
                pub2ref['CitedPublicationId'] = [pub2int_map(pid) for pid in pub2ref['CitedPublicationId'].values]

                with gzip.open(os.path.join(self.path2database, 'pub2int.json.gz'), 'w') as outfile:
                        outfile.write(json.dumps(pub2int).encode('utf8'))

            if preprocess:
                fname = os.path.join(self.path2database, self.path2pub2ref, '{}{}.{}'.format(self.path2pub2ref, 0, self.database_extension))
                self.save_data_file(pub2ref, fname, key =self.path2pub2ref)

            return pub2ref

        else:
            raise FileNotFoundError('The archive {0} does not contain a citation file: {1}.'.format(archive_name, 'aps-dataset-citations'))

    def parse_publicationauthoraffiliation(self, preprocess = False, num_file_lines=10**7, show_progress=False):
        raise NotImplementedError("APS is stored as a json archive.  Run preprocess to parse the archive.")

    def parse_fields(self, preprocess = False, num_file_lines=10**7, show_progress=False):
        raise NotImplementedError("APS is stored as a json archive.  Run preprocess to parse the archive.")

    def load_journals(self, preprocess = True, columns = None, filter_dict = {}, duplicate_subset = None,
        duplicate_keep = 'last', dropna = None, prefunc2apply=None, postfunc2apply=None, show_progress=False):
    
        raise NotImplementedError("The APS does not have prespecified journal information.")




# -*- coding: utf-8 -*-
"""
.. module:: all
    :synopsis: easy interface to all of pyscisci

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

class RangeFilter():

	def __init__(self, field, min_value=None, max_value=None):

		self.field = field
		self.min = min_value
		self.max = max_value

		self.check_value = lambda s,x: False

		if not self.min is None and not self.max is None:
			self.check_value = self.fullrange

		elif not self.min is None and self.max is None:
			self.check_value = self.lowerbound

		elif self.min is None and not self.max is None:
			self.check_value = self.upperbound

		else:
			raise ValueError("One of min_value or max_value must be set.")


	def fullrange(self, value):
		return (value >= self.min) and (value <= self.max)

	def lowerbound(self, value):
		return (value >= self.min)

	def upperbound(self, value):
		return (value <= self.max)

class SetFilter():

	def __init__(self, field, value_set=None):

		self.field = field
		self.value_set = set(value_set)

	def check_value(self, value):
		return value in self.value_set

class YearFilter(RangeFilter):

	def __init__(self, min_year=None, max_year=None):

		self.field = 'Year'
		self.min = min_year
		self.max = max_year

		self.check_value = lambda s,x: False

		if not self.min is None and not self.max is None:
			self.check_value = self.fullrange

		elif not self.min is None and self.max is None:
			self.check_value = self.lowerbound

		elif self.min is None and not self.max is None:
			self.check_value = self.upperbound

class DocTypeFilter(SetFilter):

	def __init__(self, doctypes=[]):

		self.field = 'DocType'
		self.value_set=set(doctypes)

class FieldFilter(SetFilter):

	def __init__(self, valid_fields=[]):
		
		self.field = 'FieldId'
		self.value_set=set(valid_fields)

class JournalFilter(SetFilter):

	def __init__(self, valid_journals=[]):
		
		self.field = 'JournalId'
		self.value_set=set(valid_journals)





# -*- coding: utf-8 -*-
"""
.. module:: diffusionscientificcredit
    :synopsis: Rank authors based on the pagerank within their citation graph.

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import pandas as pd
import numpy as np

from pyscisci.utils import isin_sorted, groupby_count, groupby_total
from pyscisci.network import cocitation_network
from pyscisci.sparsenetworkutils import dataframe2bipartite, sparse_pagerank_scipy

def diffusion_of_scientific_credit(pub2ref, pub2author, pub=None, alpha = 0.9, max_iter = 100, tol = 1.0e-10):
    """
    Calculate the diffusion of scientific credits for each author based on :cite:`Radicchi2009authorpagerank`.

    Parameters
    ----------

    :param pub2ref : DataFrame
        A DataFrame with the citation information for each Publication.

    :param pub2author : DataFrame
        A DataFrame with the author information for each Publication.

    :param pub : DataFrame
        A DataFrame with the publication information for each Publication.

    :param alpha : float, default 0.9
        The PageRank reset probility

    :param max_iter : int, default 100
        The maximum number of iterations when appllying the power method.

    :param tol : float, default 1.0e-10
        The error tolerance when appllying the power method.

    Returns
    -------
    credit_share, numpy array
        If temporal == False:
            The adjacency matrix for the co-citation network

        If temporal == True:
            A dictionary with key for each year, and value of the adjacency matrix for the cocitation network induced
            by citing publications in that year.

    author2int, dict
        A mapping of the AuthorIds from the focus publication to the column of the credit share vector or matrix (see above).

    """

    """
    Diffusion of Scientific Credits and the Ranking of Scientists
    Radicchi et al (2009) Phys Rev E


    author_subset - each row is one article & author combination.  
            at least two columns 'name':author name, and 'teamsize':number of total authors on the paper

    full_citation - each row is one citation from one article & author combination
            at least four columns 'name_citing': name of citing author, 'name_cited': name of cited author, 
            'teamsize_citing': number of authors for citing paper, 'teamsize_cited': number of authors for cited papers.
    """

    # relabel the authors to map to network nodes
    focus_authors = np.sort(pub2author['AuthorId'].unique())
    author2int = {aid:i for i, aid in enumerate(focus_authors)}
    Nauthors = len(author2int)

    pub2author.drop_duplicates(subset=['PublicationId', 'AuthorId'], inplace=True)
    pub2author['AuthorId'] = [author2int.get(aid, None) for aid in pub2author['AuthorId'].values]

    # check if we are given the teamsize in publication information
    if (not pub is None) and 'TeamSize' in list(pub):
        teamsize = {pid:ts for pid, ts in pub[['PublicationId', 'TeamSize']].values}
    
    # otherwies we need to calculate teamsize based on the authorship information
    else:
        teamsize = pub2author.groupby('PublicationId')['AuthorId'].nunique()


    full_citation = pub2ref.merge(pub2author[['PublicationId', 'AuthorId']], left_on = 'CitingPublicationId', right_on = 'PublicationId')
    del full_citation['PublicationId']
    full_citation.rename(columns={'AuthorId':'CitingAuthorId'}, inplace=True)

    full_citation = full_citation.merge(pub2author[['PublicationId', 'AuthorId']], left_on = 'CitedPublicationId', right_on = 'PublicationId')
    del full_citation['PublicationId']
    full_citation.rename(columns={'AuthorId':'CitedAuthorId'}, inplace=True)

    full_citation.dropna(inplace=True)

    # now add in the teamsize information to make edge weights
    full_citation['edge_weight'] = [1.0/(teamsize.get(citing_pid, 1) * teamsize.get(cited_pid, 1)) for citing_pid, cited_pid in full_citation[['CitingPublicationId', 'CitedPublicationId']].values]


    adj_mat = dataframe2bipartite(full_citation, rowname='CitingAuthorId', colname='CitedAuthorId', 
        shape = (Nauthors,Nauthors), weightname = 'edge_weight')


    # make the weighted productivity vector to intialize the pagerank
    pub2author['AuthorCredit'] = [1/teamsize.get(pid, 1) for pid in pub2author['PublicationId'].values]
    weighted_productivity = groupby_total(pub2author, colgroupby = 'AuthorId', colcountby = 'AuthorCredit').sort_values('AuthorId')
    # norm vector
    weighted_productivity['AuthorCreditTotal'] = weighted_productivity['AuthorCreditTotal'] / weighted_productivity['AuthorCreditTotal'].sum()

    # run the power method to solve the diffusion 
    sc = sparse_pagerank_scipy(adj_mat, alpha= alpha, 
        personalization=weighted_productivity['AuthorCreditTotal'].values, 
        initialization=weighted_productivity['AuthorCreditTotal'].values,
                   max_iter=max_iter, tol=tol, dangling=None)

    return sc, author2int



# -*- coding: utf-8 -*-
"""
.. module:: scinlp
    :synopsis: The Natual Langauge Processing for SciSci

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>

 """
import os
import sys
import re
from collections import defaultdict, Counter
import pandas as pd
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer

from pyscisci.sparsenetworkutils import project_bipartite_mat, threshold_network
from pyscisci.utils import *

import unicodedata
from unidecode import  unidecode
from nameparser import HumanName

# For now, I dont think we need to make the full pySciSci package dependent on these packages
try:
    from Levenshtein import ratio
except ImportError:
    raise ImportError('Please install python-Levenshtein to take full advantage of the NLP tools. \n pip install python-Levenshtein')

try:
    from sparse_dot_topn import awesome_cossim_topn
except ImportError:
    raise ImportError("Please install sparse_dot_topn to take full advantage of the NLP tools. \n pip install sparse_dot_topn")

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

def abstractindex2text(abstract_index):
    """
    The abstracts in OpenAlex and MAG are saved in the inverted index format.  This function coverts back to the original text.
    """

    # unpack the index list
    word_index = [ (word, idx) for word, idxlist in abstract_index.items() for idx in idxlist] 
    # sort the words
    word_index = sorted(word_index, key = lambda x : x[1])
    # insert the spaces back in and return
    return " ".join(list(zip(*word_index))[0])


def standardize_doi(doistr):
    """
    
    """

    

    
def strip_accents(text):
    try:
        text = unicode(text, 'utf-8')
    except (TypeError, NameError): # unicode is a default on python 3
        pass
    text = unicodedata.normalize('NFD', text)
    text = text.encode('ascii', 'ignore')
    text = text.decode("utf-8")
    return str(text)

def clean_names(name, remove_parentheses=True):
    name = strip_accents(name)

    if remove_parentheses:
        name = re.sub("[\(\[].*?[\)\]]", "", name) # remove text between () and []
        if '(' in name:
            name=name[:name.index('(')]
        for c in [')', ']','.']:
            name=name.replace(c, '')
    for c in [',', ' - ', '- ', ' -']:
        name=name.replace(c, '-')
    name = name.replace(' & ', '&')
    return name.strip()

def levenshtein_best_match(s1, options, lower_bound=0.75):
    """
    Uses Levenshtein distance to pick the best match from a list of options.

    Returns tuple: (index in the options list, Levenshtein dist)
    """

    return max(zip(range(len(options)), [ratio(s1, s) for s in options]), key = lambda x: x[1])

def align_publications(df1, df2=None, columns2match_exact=['Year'], column2match_approx='Title', ntop=1, cosine_lower_bound=0.75,
    use_threads=False, n_jobs=2, lev_lower_bound=0.9, show_progress=False):
    """
    Fast way to match publications between two datasets.  We first match subsets of exact values
    between the two DataFrames, as specified by `columns2match_exact`.
    We then use a fast approximate string matching to match values in `columns2match_approx` to within a threshold.

    Parameters
    ----------
    :param df1 : DataFrame
        A DataFrame with the publication information.

    :param df2 : DataFrame, Optional
        Another DataFrame with the publication information.  If None, then df1 is used again.

    :param columns2match_exact : list, Default: ['Year']
        The columns to match exactly between DataFrames.

    :param column2match_approx : list, Default: 'Title'
        The column to match approximately between DataFrames.

    :param ntop : int, Default 1
        The number of best matches from df2 to return for each row of df1.

    :param lower_bound : float, Default 0.75
        The lowerbound for cosine similarity when doing a fuzzy string match.

    :param use_threads : bool, Default False
        Use multithreading when calculating cosine similarity for fuzzy string matching.

    :param n_jobs : int, Optional, Default 2
        If use_threads is True, the number of threads to use in the parall calculation.

    :param show_progress : bool, Default False
        If True, show a progress bar tracking the calculation.

    """
    # we can do an exact match from merge
    if (columns2match_exact is None or len(columns2match_exact) > 0) and (column2match_approx is None or len(column2match_approx) == 0):
        # get the index name and reset the index to force it as a column
        indexcol = df2.index.name
        df2 = df2.reset_index(drop=False)
        # now merge the dataframes and drop duplicates giving an exact match
        mdf = df1.merge(df2[columns2match_exact + [indexcol]], how='left', on=columns2match_exact)
        mdf.drop_duplicates(subset=columns2match_exact, keep='first', inplace=True)
        return mdf[indexcol]

    # otherwise, if there is a column to match approximately then we need to prep for fuzzy matching
    elif len(column2match_approx) > 0:

        # we take a two-step approach to fuzzy matching
        # 1) first we employ a super fast but not very accurate cosine-similarity
        #    matching to narrow down the possible pair-wise matches
        #    for each string, we create feature vectors from 3-char counts
        tfidf = TfidfVectorizer(min_df=1, ngram_range = (3,3), analyzer='char', lowercase=False)
        tfidf1 = tfidf.fit_transform(df1[column2match_approx])

        matches = np.empty(tfidf1.shape[0])
        matches[:] = np.NaN

        # if there are no columns to match exactly
        if (columns2match_exact is None or len(columns2match_exact) == 0):

            tfidf2 = tfidf.transform(df2[column2match_approx])

            # 1) first do the all-to-all cosine similarity and extract up to the ntop best possible matches
            co= awesome_cossim_topn(tfidf1, tfidf2.T, ntop=ntop, lower_bound=cosine_lower_bound, use_threads=use_threads, n_jobs=n_jobs).tocoo()

            # 2) now use the Levenshtein
            for row in tqdm(set(co.row), desc="Align Publications", disable=not show_progress):
                rowcol = co.col[co.row==row]
                argmatch, lev_dist = levenshtein_best_match(df1.loc[row, column2match_approx], df2.iloc[rowcol][column2match_approx])
                if lev_dist >= lev_lower_bound:
                    matches[row] = rowcol[argmatch]


        else:

            df2groups = df2.groupby(columns2match_exact)

            def subgroup_match(subdf):
                if not df2groups.indices.get(subdf.name, None) is None:
                    sub_tfidf1 = tfidf1[subdf.index.values]

                    sub_tfidf2 = tfidf.transform(df2.loc[df2groups.indices[subdf.name]][column2match_approx])

                    #sub_tfidf2 = tfidf2[df2groups.indices[subdf.name]]
                    co = awesome_cossim_topn(sub_tfidf1, sub_tfidf2.transpose(), ntop=ntop, lower_bound=cosine_lower_bound, use_threads=use_threads, n_jobs=n_jobs).tocoo()

                    # 2) now use the Levenshtein distance to find the best match
                    for row in set(co.row):
                        rowcol = co.col[co.row==row]
                        argmatch, lev_dist = levenshtein_best_match(subdf.iloc[row][column2match_approx],
                            df2.iloc[df2groups.indices[subdf.name][rowcol]][column2match_approx])
                        if lev_dist >= lev_lower_bound:
                            matches[subdf.index.values[row]] = df2groups.indices[subdf.name][rowcol[argmatch]]

            # register our pandas apply with tqdm for a progress bar
            tqdm.pandas(desc='Publication Matches', disable= not show_progress)

            df1.groupby(columns2match_exact, group_keys=True).progress_apply(subgroup_match)

        return matches

def coword_network(df, text_column='Title', stop_words= 'english', strip_accents='ascii', lowercase=True, threshold=1, vocabulary=None, show_progress=False):
    """
    Create the co-word network.

    Parameters
    ----------
    df : DataFrame
        A DataFrame with the text snippits (typically titles or abstracts).

    text_column : str, default "Title"
        The column of text snippits

    stop_words : str or list of str, default 'english'
        The stopword dictionary to employ.
            - 'english' : sklearn english stopwords
            - list of str : a list specifying the exact stopwords to remove
            - None : no stopword dictionary (use all stopwords)

    strip_accents : str, default 'ascii'
        Remove accents and perform other character normalization during the preprocessing step.  None does nothing.

    lowercase : bool, default True
        Convert all characters to lowercase before tokenizing.

    threshold : int, default 1
        The minimum number of times two words should appear together.

    vocabulary : list, default None
        List of focus terms to limit. If not given, a vocabulary is determined from the input text. 

    show_progress : bool, default False
        If True, show a progress bar tracking the calculation.

 

    Returns
    -------
    coo_matrix or dict of coo_matrix
        The adjacency matrix for the co-authorship network

    word2int, dict
        A mapping of words to the row/column of the adjacency matrix.


    """
    required_columns = ['PublicationId', text_column]
    check4columns(df, required_columns)
    df = df[required_columns].dropna()

    df.drop_duplicates(subset=required_columns, inplace=True)

    cv = CountVectorizer(stop_words=stop_words, analyzer='word', min_df=threshold, lowercase=lowercase, strip_accents=strip_accents,
        vocabulary=vocabulary)
    bipartite_adj = cv.fit_transform(df[text_column].values)

    word2int = cv.vocabulary_

    adj_mat = project_bipartite_mat(bipartite_adj, project_to = 'col')
    adj_mat = threshold_network(adj_mat)

    return adj_mat, word2int

